from abc import ABC, abstractmethod
from .action_set import ActionSet
from typing import Union
import numpy as np
import pandas as pd

# class Agent(ABC):
#     """
#     Defines a single agent playing within the environment.
#
#     Some relevant parameters will include:
#     - Agent's personal info
#     - Percentage of adaptation towards counterfactual
#
#     Attributes:
#     - status: exited, out of or included in population
#     - outcome: None, favorable, adverse
#     """
#
#     def __init__(self):
#         pass
#
#     def update(self):
#         """Moves agent to the next timestep"""
#         pass


class BasePopulation(ABC):
    """
    Defines the population of agents within the environment.

    A population consists of a set of Agents.

    Actionable recourse supports only binary categorical features.

    This is intended to be a more "static" object, which can be manipulated
    via an Environment object.

    List of features to implement:
    - [x] Agents' personal info (data)
    - [x] categorical features
    - [x] immutable features
    - [x] step direction
    - [x] Definition the action set internally
    - [x] Amount of new Agents per update
    - [] Amount of Agents with an adverse outcome that give up (leave the
      Population)
    - [] Generator (for new agents)
    - [] Population objects should be iterable. Selecting one element should
         return an Agent.
    - [] Allow entering data as numpy array (?)
    """

    def __init__(
        self,
        data: pd.DataFrame,
        y_desired: Union[int, str] = 1,
        categorical: Union[list, np.ndarray] = None,
        immutable: Union[list, np.ndarray] = None,
        step_direction: dict = None,
    ):
        self.data = data
        self.y_desired = y_desired
        self.categorical = categorical
        self.immutable = immutable
        self.step_direction = step_direction

        self.set_actions()

    def set_actions(self, action_set=None):
        """
        To be configured with the ActionSet object from the
        ``actionable-recourse`` library.
        """

        categorical = [] if self.categorical is None else self.categorical
        immutable = [] if self.immutable is None else self.immutable
        step = {} if self.step_direction is None else self.step_direction

        if action_set is None:
            action_set = ActionSet(
                X=self.data, y_desired=self.y_desired, default_bounds=(0, 1)
            )
            for col in self.data.columns:
                if col in immutable:
                    action_set[col].actionable = False

                if col in step.keys():
                    action_set[col].step_direction = self.step_direction[col]

                if col in categorical:
                    action_set[col].variable_type = int
                else:
                    action_set[col].ub = self.data[col].max()
                    action_set[col].lb = self.data[col].min()

        self.action_set_ = action_set

        return self

    def set_params(self, **kwargs):
        """
        This should be used to add/update parameters. Use same approach as
        sklearn's.
        """
        pass



from .environments import (
    ClosedEnvironment,
    BankLoanApplication1,
    BankLoanApplication2,
    BankLoanApplication3,
    WillingnessEnvironment,
)

__all__ = [
    "ClosedEnvironment",
    "BankLoanApplication1",
    "BankLoanApplication2",
    "BankLoanApplication3",
    "WillingnessEnvironment",
]



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.neural_network import MLPRegressor

from .swarmplot import swarm


class EnvironmentPlot:
    """
    TODO: Add documentation.
    """

    _favorable = "green"
    _unfavorable = "blue"
    _previous = "red"

    def __init__(self, environment, random_state=None):
        self.environment = environment
        self.random_state = random_state

    def fit(self, X=None):
        if X is None:
            X = self.environment.population.data

        self.is_large_dim = X.shape[-1] > 2

        if self.is_large_dim:
            self._autoencoder = MLPRegressor(
                hidden_layer_sizes=(2,),
                activation="identity",
                max_iter=1000,
                random_state=self.random_state,
            )
            self._autoencoder.fit(X, X)

            X = pd.DataFrame(
                (
                    np.dot(X.values, self._autoencoder.coefs_[0])
                    + self._autoencoder.intercepts_[0]
                ),
                index=X.index,
                columns=["Component 1", "Component 2"],
            )

        self._min_x = X.min(0)
        self._max_x = X.max(0)

        return self

    def _create_mesh_grid(self, X_curr, mesh_size):
        if self.is_large_dim:
            X_ = pd.DataFrame(
                (
                    np.dot(X_curr.values, self._autoencoder.coefs_[0])
                    + self._autoencoder.intercepts_[0]
                ),
                index=X_curr.index,
                columns=["Component 1", "Component 2"],
            )
        else:
            X_ = X_curr.copy()
        _min_x_curr = X_.min(0)
        _max_x_curr = X_.max(0)

        mins = np.array([_min_x_curr, self._min_x]).min(0)
        maxs = np.array([_max_x_curr, self._max_x]).max(0)

        self.mesh_x, self.mesh_y = [
            np.array(
                np.linspace(
                    mins[i] - 0.1 * (maxs[i] - mins[i]),
                    maxs[i] + 0.1 * (maxs[i] - mins[i]),
                    mesh_size[i],
                )
            )
            for i in range(2)
        ]
        mesh = np.stack(np.meshgrid(self.mesh_x, self.mesh_y), axis=2)
        mesh = np.reshape(mesh, (mesh_size[0] * mesh_size[1], 2))

        if self.is_large_dim:
            mesh = (
                np.dot(mesh, self._autoencoder.coefs_[1])
                + self._autoencoder.intercepts_[1]
            )

        mesh = self.environment.model_.predict_proba(
            pd.DataFrame(mesh, columns=X_curr.columns)
        )[:, -1]
        self.mesh_ = np.reshape(mesh, mesh_size)
        return self

    def scatter(
        self, step=None, mesh_size=(100, 100), legend=True, title=True, ax=None
    ):
        """Visualize the population in a 2d-scatter plot."""
        if not hasattr(self, "_autoencoder"):
            self.fit()

        if step is None:
            step = self.environment.step_

        df = self.environment.metadata_[step]["population"].data
        if step > 0:
            df_prev = self.environment.metadata_[step - 1]["population"].data

        outcome = self.environment.predict(step=step).astype(bool)

        self._create_mesh_grid(df, mesh_size)

        # Project data into 2 dimensions
        if self.is_large_dim:
            df = pd.DataFrame(
                (
                    np.dot(df.values, self._autoencoder.coefs_[0])
                    + self._autoencoder.intercepts_[0]
                ),
                index=df.index,
                columns=["Component 1", "Component 2"],
            )
            df_prev = (
                pd.DataFrame(
                    (
                        np.dot(df_prev.values, self._autoencoder.coefs_[0])
                        + self._autoencoder.intercepts_[0]
                    ),
                    index=df_prev.index,
                    columns=["Component 1", "Component 2"],
                )
                if step > 0
                else None
            )

        # Visualize probabilities
        if ax is None:
            _, ax = plt.subplots(1, 1, figsize=(10, 10))

        prob = self.mesh_

        ax.contourf(self.mesh_x, self.mesh_y, prob, levels=100, alpha=0.5, cmap="Blues")

        # Visualize agents
        if step > 0:
            move = self.environment._get_moving_agents(step)

            # Plot movement lines
            ax.plot(
                [df_prev.loc[move].iloc[:, 0], df.loc[move].iloc[:, 0]],
                [df_prev.loc[move].iloc[:, 1], df.loc[move].iloc[:, 1]],
                color="gray",
                alpha=0.5,
                zorder=1,
            )

            # Plot previous location for moving agents
            ax.scatter(
                x=df_prev.loc[move].iloc[:, 0],
                y=df_prev.loc[move].iloc[:, 1],
                alpha=0.3,
                color=self._previous,
                label="Prev. position" if legend else None,
            )

        # Plot all current agents with unfavorable outcome
        ax.scatter(
            x=df.iloc[~outcome.values, 0],
            y=df.iloc[~outcome.values, 1],
            color=self._unfavorable,
            alpha=0.5,
            label="Unfavorable" if legend else None,
        )

        # Plot all current agents with favorable outcome
        ax.scatter(
            x=df.iloc[outcome.values, 0],
            y=df.iloc[outcome.values, 1],
            color=self._favorable,
            alpha=0.5,
            label="Favorable" if legend else None,
        )
        if legend:
            ax.legend()
        if title:
            ax.set_title(f"Population at t={step}")
        ax.set_xlabel(df.columns[0])
        ax.set_ylabel(df.columns[1])
        return ax

    def agent_scores(
        self, min_step=None, max_step=None, legend=True, title=True, ax=None
    ):
        """Visualize population scores across multiple time steps."""
        if not hasattr(self, "_autoencoder"):
            self.fit()

        df_list = [
            (i, metadata["population"].data, metadata["threshold"])
            for i, metadata in self.environment.metadata_.items()
        ][min_step:max_step]

        if ax is None:
            _, ax = plt.subplots(1, 1, figsize=(10, 10))

        for step, df, threshold in df_list:
            prob = self.environment.model_.predict_proba(df)[:, -1]
            outcome = self.environment.predict(step=step).astype(bool)

            # Set up x coordinates to form swarm plots
            x = np.ones(df.shape[0], dtype=int) * step
            x = x + swarm(prob)

            add_labels = legend and step in [min_step, 0]

            ax.scatter(
                x=x[~outcome.values],
                y=prob[~outcome.values],
                color=self._unfavorable,
                alpha=0.5,
                label=("Unfavorable" if add_labels else None),
            )
            ax.scatter(
                x=x[outcome.values],
                y=prob[outcome.values],
                color=self._favorable,
                alpha=0.5,
                label=("Favorable" if add_labels else None),
            )

        ax.plot(
            [meta[0] for meta in df_list],
            [meta[2] for meta in df_list],
            color=self._previous,
            label="Threshold",
            alpha=0.3,
        )

        if title:
            ax.set_title(f"Population scores over {len(df_list)} time steps")
        if legend:
            ax.legend()
        ax.set_xticks([meta[0] for meta in df_list])
        ax.set_xlabel("Step")
        ax.set_ylabel("Score")

        return ax

    def scores_histogram(self, step=None, **kwargs):
        """
        Plots the histogram of the scores at ``step``.

        See documentation for plt.hist in order to pass kwargs
        """
        if step is None:
            step = self.environment.step_

        return plt.hist(
            self.environment.model_.predict_proba(
                self.environment.metadata_[step]["population"].data
            )[:, -1],
            **kwargs,
        )

    def scores_kde(self, step=None, **kwargs):
        """
        Plots the KDE of the scores at ``step`` using Gaussian kernels.

        See documentation for pandas.DataFrame.plot.kde in order to pass kwargs
        """
        if step is None:
            step = self.environment.step_

        return pd.Series(
            self.environment.model_.predict_proba(
                self.environment.metadata_[step]["population"].data
            )[:, -1]
        ).plot.kde(**kwargs)

    def population_size(
        self,
        min_step=None,
        max_step=None,
        legend=True,
        title=True,
        stacked=True,
        ax=None,
        **kwargs,
    ):
        """
        plots the number of agents per time step

        See documentation for pandas.DataFrame.plot.bar in order to pass kwargs
        """

        pop_list = [
            (i, metadata["population"].data.shape[0], metadata["outcome"].sum())
            for i, metadata in self.environment.metadata_.items()
        ][min_step:max_step]

        df = pd.DataFrame(pop_list, columns=["step", "pop", "favorable"])
        df.set_index("step", inplace=True)
        df["unfavorable"] = df["pop"] - df["favorable"]
        df.drop(columns="pop", inplace=True)
        ax = df[["unfavorable", "favorable"]].plot.bar(stacked=stacked, ax=ax, **kwargs)

        if title:
            ax.set_title(f"Population size over {len(pop_list)} time steps")
        if legend:
            ax.legend()
        ax.set_xlabel("Step")
        ax.set_ylabel("Number of agents")

        return ax



import numpy as np


def swarm(y, nbins=None):
    """
    Returns x coordinates for the points in ``y``, so that plotting ``x`` and
    ``y`` results in a swarm plot.
    """
    y = np.asarray(y)
    if nbins is None:
        nbins = len(y) // 6

    # Get upper bounds of bins
    x = np.zeros(len(y))
    ylo = np.min(y)
    yhi = np.max(y)
    dy = (yhi - ylo) / nbins
    ybins = np.linspace(ylo + dy, yhi - dy, nbins - 1)

    # Divide indices into bins
    i = np.arange(len(y))
    ibs = [0] * nbins
    ybs = [0] * nbins
    nmax = 0
    for j, ybin in enumerate(ybins):
        f = y <= ybin
        ibs[j], ybs[j] = i[f], y[f]
        nmax = max(nmax, len(ibs[j]))
        f = ~f
        i, y = i[f], y[f]
    ibs[-1], ybs[-1] = i, y
    nmax = max(nmax, len(ibs[-1]))

    # Assign x indices
    dx = 1 / (nmax // 2)
    for i, y in zip(ibs, ybs):
        if len(i) > 1:
            j = len(i) % 2
            i = i[np.argsort(y)]
            a = i[j::2]
            b = i[j + 1 :: 2]
            x[a] = (0.5 + j / 3 + np.arange(len(b))) * dx
            x[b] = (0.5 + j / 3 + np.arange(len(b))) * -dx

    x = ((x - x.min()) / (x.max() - x.min())) * 0.8 - 0.4

    return x



import warnings
import numpy as np
from .base import BaseRecourse


class NFeatureRecourse(BaseRecourse):
    def __init__(self, model, n_features=None, threshold=0.5):
        self.model = model
        self.n_features = n_features
        self.threshold = threshold

    def _counterfactual(self, agent, action_set):
        agent = agent.copy()

        # Do not change if the agent is over the threshold
        if self.model.predict_proba(agent.to_frame().T)[0, -1] >= self.threshold:
            return agent

        intercept, coefficients, model = self._get_coefficients()

        # Get base vector
        base_vector = coefficients.copy().squeeze()
        n_features = (
            base_vector.shape[0] if self.n_features is None else self.n_features
        )

        is_usable = np.array(
            [
                action_set[col].step_direction in [np.sign(coeff), 0]
                and action_set[col].actionable
                for col, coeff in zip(agent.index, base_vector)
            ]
        )
        base_vector[~is_usable] = 0

        # Use features with highest contribution towards the threshold
        rejected_features = np.argsort(np.abs(base_vector))[:-n_features]
        base_vector[rejected_features] = 0

        base_vector = base_vector / np.linalg.norm(base_vector)
        multiplier = (-intercept - np.dot(agent.values, coefficients.T)) / np.dot(
            base_vector, coefficients.T
        )
        counterfactual = agent + multiplier * base_vector

        # Check if base_vector adjustments are not generating invalid counterfactuals
        for i in range(agent.shape[0]):
            # Adjust vector according to features' bounds
            lb, ub = np.array(action_set.lb), np.array(action_set.ub)
            lb_valid = counterfactual >= lb
            ub_valid = counterfactual <= ub

            if lb_valid.all() and ub_valid.all():
                break

            if not lb_valid.all():
                # Fix values to its lower bound
                idx = np.where(~lb_valid)[0]
                agent[idx] = lb[idx]
                base_vector[idx] = 0

            if not ub_valid.all():
                # Fix values to its upper bound
                idx = np.where(~ub_valid)[0]
                agent[idx] = ub[idx]
                base_vector[idx] = 0

            if (base_vector == 0).all():
                # All max/min boundaries have been met.
                counterfactual = agent
            else:
                # Redefine counterfactual after adjusting the base vector
                base_vector = base_vector / np.linalg.norm(base_vector)
                multiplier = (
                    -intercept - np.dot(agent.values, coefficients.T)
                ) / np.dot(base_vector, coefficients.T)
                counterfactual = agent + multiplier * base_vector

        lb, ub = np.array(action_set.lb), np.array(action_set.ub)
        lb_valid = counterfactual >= lb
        ub_valid = counterfactual <= ub
        if not (lb_valid.all() and ub_valid.all()):
            warnings.warn(
                "Could not generate a counterfactual to reach the desired threshold."
            )

        return counterfactual



# This file contains convenience functions for CPLEX MIP Objects
import numpy as np
from functools import reduce
from cplex import Cplex, SparsePair
from cplex.exceptions import CplexError


# Copying
def copy_cplex(cpx):
    """
    Copy a Cplex object
    :param cpx: Cplex object
    :return: Copy of Cplex object
    """
    cpx_copy = Cplex(cpx)
    cpx_parameters = cpx.parameters.get_changed()
    for pname, pvalue in cpx_parameters:
        phandle = reduce(getattr, str(pname).split("."), cpx_copy)
        phandle.set(pvalue)
    return cpx_copy


# Building
def add_variable_cpx(cpx, name, obj, ub, lb, vtype):
    """
    Convenience function to add a variable to a Cplex() object
    :param cpx: handle to Cplex() object
    :param name: variable name
    :param obj: coefficient in linear objective
    :param ub: upper bound on variable
    :param lb: lower bound on variable
    :param vtype: variable type
    :return:
    """

    # name
    if isinstance(name, np.ndarray):
        name = name.tolist()
    elif isinstance(name, str):
        name = [name]

    nvars = len(name)

    # convert inputs
    if nvars == 1:
        # convert to list
        name = name if isinstance(name, list) else [name]
        obj = [float(obj[0])] if isinstance(obj, list) else [float(obj)]
        ub = [float(ub[0])] if isinstance(ub, list) else [float(ub)]
        lb = [float(lb[0])] if isinstance(lb, list) else [float(lb)]
        vtype = vtype if isinstance(vtype, list) else [vtype]

    else:
        # convert to list
        if isinstance(vtype, np.ndarray):
            vtype = vtype.tolist()
        elif isinstance(vtype, str):
            if len(vtype) == 1:
                vtype = nvars * [vtype]
            elif len(vtype) == nvars:
                vtype = list(vtype)
            else:
                raise ValueError(
                    "invalid length: len(vtype) = %d. expected either 1 or %d"
                    % (len(vtype), nvars)
                )

        if isinstance(obj, np.ndarray):
            obj = obj.tolist()
        elif isinstance(obj, list):
            if len(obj) == nvars:
                obj = [float(v) for v in obj]
            elif len(obj) == 1:
                obj = nvars * [float(obj)]
            else:
                raise ValueError(
                    "invalid length: len(obj) = %d. expected either 1 or %d"
                    % (len(obj), nvars)
                )
        else:
            obj = nvars * [float(obj)]

        if isinstance(ub, np.ndarray):
            ub = ub.tolist()
        elif isinstance(ub, list):
            if len(ub) == nvars:
                ub = [float(v) for v in ub]
            elif len(ub) == 1:
                ub = nvars * [float(ub)]
            else:
                raise ValueError(
                    "invalid length: len(ub) = %d. expected either 1 or %d"
                    % (len(ub), nvars)
                )
        else:
            ub = nvars * [float(ub)]

        if isinstance(lb, np.ndarray):
            lb = lb.tolist()
        elif isinstance(lb, list):
            if len(lb) == nvars:
                lb = [float(v) for v in lb]
            elif len(ub) == 1:
                lb = nvars * [float(lb)]
            else:
                raise ValueError(
                    "invalid length: len(lb) = %d. expected either 1 or %d"
                    % (len(lb), nvars)
                )
        else:
            lb = nvars * [float(lb)]

    # check that all components are lists
    assert isinstance(name, list)
    assert isinstance(obj, list)
    assert isinstance(ub, list)
    assert isinstance(lb, list)
    assert isinstance(vtype, list)

    # check components
    for n in range(nvars):
        assert isinstance(name[n], str)
        assert isinstance(obj[n], float)
        assert isinstance(ub[n], float)
        assert isinstance(lb[n], float)
        assert isinstance(vtype[n], str)

    if (vtype.count(vtype[0]) == len(vtype)) and vtype[0] == cpx.variables.type.binary:
        cpx.variables.add(names=name, obj=obj, types=vtype)
    else:
        cpx.variables.add(names=name, obj=obj, ub=ub, lb=lb, types=vtype)


# Parameter Setting
DEFAULT_CPLEX_PARAMETERS = {
    #
    "display_cplex_progress": False,
    # set to True to show CPLEX progress in console
    #
    "n_cores": 1,
    # Number of CPU cores to use in B & B
    # May have to set n_cores = 1 in order to use certain control callbacks in CPLEX 12.7.0 and earlier
    #
    "randomseed": 0,
    # This parameter sets the random seed differently for diversity of solutions.
    # https://www.ibm.com/support/knowledgecenter/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/RandomSeed.html
    #
    "time_limit": 1e75,
    # runtime before stopping,
    #
    "node_limit": 9223372036800000000,
    # number of nodes to process before stopping,
    #
    "mipgap": np.finfo("float").eps,
    # Sets a relative tolerance on the gap between the best integer objective and the objective of the best node remaining.
    # https://www.ibm.com/support/knowledgecenter/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/EpGap.html
    #
    "absmipgap": np.finfo("float").eps,
    # Sets an absolute tolerance on the gap between the best integer objective and the objective of the best node remaining.
    # When this difference falls below the value of this parameter, the mixed integer optimization is stopped.
    # https://www.ibm.com/support/knowledgecenter/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/EpAGap.html
    #
    "objdifference": 0.0,
    # Used to update the cutoff each time a mixed integer solution is found. This value is subtracted from objective
    # value of the incumbent update, so that the solver ignore solutions that will not improve the incumbent by at
    # least this amount.
    # https://www.ibm.com/support/knowledgecenter/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/ObjDif.html#
    #
    "integrality_tolerance": 0.0,
    # specifies the amount by which an variable can differ from an integer and be considered integer feasible. 0 is OK
    # https://www.ibm.com/support/knowledgecenter/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/EpInt.html
    #
    "mipemphasis": 0,
    # Controls trade-offs between speed, feasibility, optimality, and moving bounds in MIP.
    # 0     =	Balance optimality and feasibility; default
    # 1	    =	Emphasize feasibility over optimality
    # 2	    =	Emphasize optimality over feasibility
    # 3 	=	Emphasize moving best bound
    # 4	    =	Emphasize finding hidden feasible solutions
    # https://www.ibm.com/support/knowledgecenter/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/MIPEmphasis.html
    #
    "bound_strengthening": -1,
    # Decides whether to apply bound strengthening in mixed integer programs (MIPs).
    # https://www.ibm.com/support/knowledgecenter/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/BndStrenInd.html
    # -1    = cplex chooses
    # 0     = no bound strengthening
    # 1     = bound strengthening
    #
    "cover_cuts": -1,
    # Decides whether or not cover cuts should be generated for the problem.
    # https://www.ibm.com/support/knowledgecenter/en/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/Covers.html
    # -1    = Do not generate cover cuts
    # 0	    = Automatic: let CPLEX choose
    # 1	    = Generate cover cuts moderately
    # 2	    = Generate cover cuts aggressively
    # 3     = Generate cover cuts very  aggressively
    #
    "zero_half_cuts": -1,
    # Decides whether or not to generate zero-half cuts for the problem. (set to off since these are not effective)
    # https://www.ibm.com/support/knowledgecenter/en/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/ZeroHalfCuts.html
    # -1    = Do not generate MIR cuts
    # 0	    = Automatic: let CPLEX choose
    # 1	    = Generate MIR cuts moderately
    # 2	    = Generate MIR cuts aggressively
    #
    "mir_cuts": -1,
    # Decides whether or not to generate mixed-integer rounding cuts for the problem. (set to off since these are not effective)
    # https://www.ibm.com/support/knowledgecenter/en/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/MIRCuts.html
    # -1    = Do not generate zero-half cuts
    # 0	    = Automatic: let CPLEX choose; default
    # 1	    = Generate zero-half cuts moderately
    # 2	    = Generate zero-half cuts aggressively
    #
    "implied_bound_cuts": 0,
    # Decides whether or not to generate valid implied bound cuts for the problem.
    # https://www.ibm.com/support/knowledgecenter/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/ImplBdLocal.html
    # -1    = Do not generate locally valid implied bound cuts
    # 0	    = Automatic: let CPLEX choose; default
    # 1	    = Generate locally valid implied bound cuts moderately
    # 2	    = Generate locally valid implied bound cuts aggressively
    # 3	    = Generate locally valid implied bound cuts very aggressively
    #
    "locally_implied_bound_cuts": 3,
    # Decides whether or not to generate locally valid implied bound cuts for the problem.
    # https://www.ibm.com/support/knowledgecenter/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/ImplBdLocal.html
    # -1    = Do not generate locally valid implied bound cuts
    # 0	    = Automatic: let CPLEX choose; default
    # 1	    = Generate locally valid implied bound cuts moderately
    # 2	    = Generate locally valid implied bound cuts aggressively
    # 3	    = Generate locally valid implied bound cuts very aggressively
    #
    "scale_parameters": 0,
    # Decides how to scale the problem matrix.
    # https://www.ibm.com/support/knowledgecenter/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/ScaInd.html
    # 0     = equilibration scaling
    # 1     = aggressive scaling
    # -1    = no scaling
    #
    "numerical_emphasis": 0,
    # Emphasizes precision in numerically unstable or difficult problems.
    # https://www.ibm.com/support/knowledgecenter/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/NumericalEmphasis.html
    # 0     = off
    # 1     = on
    #
    "poolsize": 100,
    # Limits the number of solutions kept in the solution pool
    # https://www.ibm.com/support/knowledgecenter/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/SolnPoolCapacity.html
    # number of feasible solutions to keep in solution pool
    #
    "poolrelgap": float("nan"),
    # Sets a relative tolerance on the objective value for the solutions in the solution pool.
    # https://www.ibm.com/support/knowledgecenter/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/SolnPoolGap.html
    #
    "poolreplace": 2,
    # Designates the strategy for replacing a solution in the solution pool when the solution pool has reached its capacity.
    # https://www.ibm.com/support/knowledgecenter/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/SolnPoolReplace.html
    # 0	= Replace the first solution (oldest) by the most recent solution; first in, first out; default
    # 1	= Replace the solution which has the worst objective
    # 2	= Replace solutions in order to build a set of diverse solutions
    #
    "repairtries": 20,
    # Limits the attempts to repair an infeasible MIP start.
    # https://www.ibm.com/support/knowledgecenter/SSSA5P_12.8.0/ilog.odms.cplex.help/CPLEX/Parameters/topics/RepairTries.html
    # -1	None: do not try to repair
    #  0	Automatic: let CPLEX choose; default
    #  N	Number of attempts
    #
    "nodefilesize": (120 * 1024) / 1,
    # size of the node file (for large scale problems)
    # if the B & B can no longer fit in memory, then CPLEX stores the B & B in a node file
}


def set_cpx_display_options(
    cpx, display_mip=True, display_parameters=False, display_lp=False
):
    """
    Convenience function to turn on/off CPLEX functions
    :param cpx:
    :param display_mip:
    :param display_parameters:
    :param display_lp:
    :return:
    """
    cpx.parameters.mip.display.set(display_mip)
    cpx.parameters.simplex.display.set(display_lp)
    cpx.parameters.paramdisplay.set(display_parameters)

    if not (display_mip or display_lp):
        cpx.set_results_stream(None)
        cpx.set_log_stream(None)
        cpx.set_error_stream(None)
        cpx.set_warning_stream(None)

    return cpx


def set_cpx_parameters(cpx, param=DEFAULT_CPLEX_PARAMETERS):
    """
    Set parameters of a Cplex object
    :param cpx: Cplex object
    :param param: dictionary of parameters
    :return: cpx
    """

    # get parameter handle
    p = cpx.parameters

    # Record calls to C API
    # cpx.parameters.record.set(True)

    if param["display_cplex_progress"] is (None or False):
        cpx = set_cpx_display_options(
            cpx, display_mip=False, display_lp=False, display_parameters=False
        )

    # major parameters
    p.randomseed.set(param["randomseed"])
    p.threads.set(param["n_cores"])
    p.output.clonelog.set(0)
    p.parallel.set(1)

    # solution strategy
    p.emphasis.mip.set(param["mipemphasis"])
    p.preprocessing.boundstrength.set(param["bound_strengthening"])

    # cuts
    p.mip.cuts.implied.set(param["implied_bound_cuts"])
    p.mip.cuts.localimplied.set(param["locally_implied_bound_cuts"])
    p.mip.cuts.zerohalfcut.set(param["zero_half_cuts"])
    p.mip.cuts.mircut.set(param["mir_cuts"])
    p.mip.cuts.covers.set(param["cover_cuts"])
    #
    # tolerances
    p.emphasis.numerical.set(param["numerical_emphasis"])
    p.mip.tolerances.integrality.set(param["integrality_tolerance"])

    # initialization
    p.mip.limits.repairtries.set(param["repairtries"])

    # solution pool
    p.mip.pool.capacity.set(param["poolsize"])
    p.mip.pool.replace.set(param["poolreplace"])

    # stopping
    p.mip.tolerances.mipgap.set(param["mipgap"])
    p.mip.tolerances.absmipgap.set(param["absmipgap"])

    if param["time_limit"] < DEFAULT_CPLEX_PARAMETERS["time_limit"]:
        cpx = set_cpx_time_limit(cpx, param["time_limit"])

    if param["node_limit"] < DEFAULT_CPLEX_PARAMETERS["node_limit"]:
        cpx = set_cpx_node_limit(cpx, param["node_limit"])

    return cpx


def get_cpx_parameters(cpx):
    """
    :param cpx: Cplex object
    :return: Cplex object
    """

    p = cpx.parameters

    param = {
        # major
        "display_cplex_progress": p.mip.display.get() > 0,
        "randomseed": p.randomseed.get(),
        "n_cores": p.threads.set.get(),
        #
        # strategy
        "mipemphasis": p.emphasis.mip.get(),
        "scale_parameters": p.read.scale.get(),
        "locally_implied_bound_cuts": p.mip.cuts.localimplied.get(),
        #
        # stopping
        "time_limit": p.mip.timelimit.get(),
        "node_limit": p.mip.limits.nodes.get(),
        "mipgap": p.mip.tolerances.mipgap.get(),
        "absmipgap": p.mip.tolerances.absmipgap.get(),
        #
        # mip tolerances
        "integrality_tolerance": p.mip.tolerances.integrality.get(),
        "numerical_emphasis": p.parameters.emphasis.numerical.get(),
        #
        # solution pool
        "repairtries": p.mip.limits.repairtries.get(),
        "poolsize": p.mip.pool.capacity.get(),
        "poolreplace": p.mip.pool.replace.get(),
        #
        # node file
        # mip.parameters.workdir.Cur  = exp_workdir;
        # mip.parameters.workmem.Cur                    = cplex_workingmem;
        # mip.parameters.mip.strategy.file.Cur          = 2; %nodefile uncompressed
        # mip.parameters.mip.limits.treememory.Cur      = cplex_nodefilesize;
    }

    return param


def set_mip_cutoff_values(cpx, objval, objval_increment):
    """
    Set the cutoff values used by the Cplex solver
    :param cpx: Cplex object
    :param objval:
    :param objval_increment:
    :return:
    """
    assert objval >= 0.0
    assert objval_increment >= 0.0
    p = cpx.parameters
    p.mip.tolerances.uppercutoff.set(float(objval))
    p.mip.tolerances.objdifference.set(0.95 * float(objval_increment))
    p.mip.tolerances.absmipgap.set(0.95 * float(objval_increment))
    return cpx


def set_cpx_time_limit(cpx, time_limit=None):
    """
    Convenience function to set a time limit on a Cplex object
    :param cpx: Cplex object
    :param time_limit: time limit in seconds
    :return: cpx: Cplex object
    """
    max_time_limit = float(cpx.parameters.timelimit.max())

    if time_limit is None:
        time_limit = max_time_limit
    else:
        time_limit = float(time_limit)
        time_limit = min(time_limit, max_time_limit)

    assert time_limit >= 0.0
    cpx.parameters.timelimit.set(time_limit)
    return cpx


def set_cpx_node_limit(cpx, node_limit=None):
    """
    Convenience function to set a node limit on a Cplex object.
    The node limit determines the maximum number of nodes that can be solved in
    branch and bound.
    :param cpx: Cplex object
    :param node_limit: time limit in seconds
    :return: cpx: Cplex object
    """
    max_node_limit = cpx.parameters.mip.limits.nodes.max()
    if node_limit == float("inf"):
        node_limit = max_node_limit
    elif node_limit is None:
        node_limit = max_node_limit
    else:
        node_limit = int(node_limit)
        node_limit = min(node_limit, max_node_limit)
    assert node_limit >= 0.0
    cpx.parameters.mip.limits.nodes.set(node_limit)
    return cpx


def toggle_cpx_preprocessing(cpx, toggle=True):
    """
    Toggle preprocessing on a Cplex object.
    This function is helpful for debugging, and running computational experiments
    :param cpx: Cplex object
    :param toggle: set to True to turn on pre-processing
    :return: Cplex object
    """
    # presolve
    # mip.parameters.preprocessing.presolve.help()
    # 0 = off
    # 1 = on

    # boundstrength
    # type of bound strengthening  :
    # -1 = automatic
    # 0 = off
    # 1 = on

    # reduce
    # mip.parameters.preprocessing.reduce.help()
    # type of primal and dual reductions  :
    # 0 = no primal and dual reductions
    # 1 = only primal reductions
    # 2 = only dual reductions
    # 3 = both primal and dual reductions

    # coeffreduce strength
    # level of coefficient reduction  :
    #   -1 = automatic
    #   0 = none
    #   1 = reduce only to integral coefficients
    #   2 = reduce any potential coefficient
    #   3 = aggressive reduction with tilting

    # dependency
    # indicator for preprocessing dependency checker  :
    #   -1 = automatic
    #   0 = off
    #   1 = at beginning
    #   2 = at end
    #   3 = at both beginning and end
    if toggle:
        cpx.parameters.preprocessing.aggregator.reset()
        cpx.parameters.preprocessing.reduce.reset()
        cpx.parameters.preprocessing.presolve.reset()
        cpx.parameters.preprocessing.coeffreduce.reset()
        cpx.parameters.preprocessing.boundstrength.reset()
    else:
        cpx.parameters.preprocessing.aggregator.set(0)
        cpx.parameters.preprocessing.reduce.set(0)
        cpx.parameters.preprocessing.presolve.set(0)
        cpx.parameters.preprocessing.coeffreduce.set(0)
        cpx.parameters.preprocessing.boundstrength.set(0)

    return cpx


# Solution Statistics
def get_stats_cpx(cpx):
    """
    Returns information associated with the current best solution for the mip
    :param cpx: Cplex object
    :return: dictionary containing information about the solution to cpx
    """

    INITIAL_SOLUTION_INFO = {
        "status": "no solution exists",
        "status_code": float("nan"),
        "has_solution": False,
        "has_mipstats": False,
        "iterations": 0,
        "nodes_processed": 0,
        "nodes_remaining": 0,
        "values": float("nan"),
        "objval": float("nan"),
        "upperbound": float("nan"),
        "lowerbound": float("nan"),
        "gap": float("nan"),
    }

    info = dict(INITIAL_SOLUTION_INFO)
    try:
        sol = cpx.solution
        progress_info = {
            "status": sol.get_status_string(),
            "status_code": sol.get_status(),
            "iterations": sol.progress.get_num_iterations(),
            "nodes_processed": sol.progress.get_num_nodes_processed(),
            "nodes_remaining": sol.progress.get_num_nodes_remaining(),
        }
        info.update(progress_info)
        info["has_mipstats"] = True
    except CplexError:
        pass

    try:
        sol = cpx.solution
        solution_info = {
            "values": np.array(sol.get_values()),
            "objval": sol.get_objective_value(),
            "upperbound": sol.MIP.get_cutoff(),
            "lowerbound": sol.MIP.get_best_objective(),
            "gap": sol.MIP.get_mip_relative_gap(),
        }
        info.update(solution_info)
        info["has_solution"] = True
    except CplexError:
        pass

    return info


# Initialization
def add_mip_start_cpx(cpx, solution, effort_level=1, name=None):
    """
    :param cpx: Cplex object
    :param solution: solution vector (list or np.array)
    :param effort_level:    (must be one of the values of mip.MIP_starts.effort_level)
                            1 <-> check_feasibility
                            2 <-> solve_fixed
                            3 <-> solve_MIP
                            4 <-> repair
                            5 <-> no_check
    :param name: name to identify the solution
    :return: Cplex object
    """
    if isinstance(solution, np.ndarray):
        solution = solution.tolist()

    mip_start = SparsePair(val=solution, ind=list(range(len(solution))))
    if name is None:
        cpx.MIP_starts.add(mip_start, effort_level)
    else:
        cpx.MIP_starts.add(mip_start, effort_level, name)

    return cpx



"""
Adapted from:
https://github.com/ustunb/actionable-recourse/blob/master/recourse/flipset.py
"""

import numpy as np
import pandas as pd
from .utils import parse_classifier_args
from ...populations import ActionSet
from .builder import RecourseBuilder


class Flipset(object):
    """
    List of actions that will flip the predicted value of a classifier from x
    """

    _valid_enumeration_types = {"mutually_exclusive", "distinct_subsets"}
    _valid_cost_types = {"total", "local", "max"}

    df_column_names = [
        "cost",
        "size",
        "features",
        "feature_idx",
        "x",
        "x_new",
        "score_new",
        "yhat_new",
        "feasible",
        "flipped",
    ]

    def __init__(self, x, action_set, solver="cplex", **kwargs):
        # attach action set
        assert isinstance(action_set, ActionSet)
        self.action_set = action_set
        self._n_variables = len(action_set)
        self._variable_names = action_set.name
        self._solver = solver

        # attach feature vector
        assert isinstance(x, (list, np.ndarray))
        self._x = np.array(x, dtype=np.float_).flatten()

        # attach coefficients
        self._coefs, self._intercept = parse_classifier_args(**kwargs)

        # initialize Flipset attributes
        self._builder = kwargs.get("builder")
        self._items = []
        self._df = pd.DataFrame(columns=Flipset.df_column_names, dtype=object)
        self._sort_args = {
            "by": ["size", "cost", "score_new"],
            "inplace": True,
            "axis": 0,
        }

    def __len__(self):
        """
        :return: # of items in the flipset
        """
        return len(self._items)

    def __str__(self):
        return str(self._df[["cost", "size", "features", "x", "x_new"]])

    def __repr__(self):
        s = [
            "Flipset with %d Items" % len(self),
            "x: %r" % self._x,
            "w: (%s)" % self._coefs,
            "items: %r" % self._items,
        ]
        return "\n".join(s)

    @property
    def x(self):
        """
        :return: feature vector
        """
        return self._x

    @property
    def solutions_info(self):
        """
        Dictionary representation of Flipset
        This is a list containing mildly processed output from recourse.builder
        """
        return self._items

    @property
    def items(self):
        return list(
            map(
                lambda x: dict(zip(self.action_set._names, x["actions"].tolist())),
                self._items,
            )
        )

    @property
    def actions(self):
        return list(map(lambda x: x["actions"], self._items))

    @property
    def df(self):
        """
        Pandas DataFrame representation of Flipset
        Each row represents a different action to flip the prediction
        Rows are sorted according to the last arguments passed to the Flipset.sort()

        DESCRIPTION OF COLUMNS
        ----------------------
        cost:           cost of action
        size:           # of altered features
        features:       names of altered variables
        feature_idx:    column indices of altered features
        x:              values of x[j] for j in feature_idx
        x_new:          values of x[j] + a[j]) for j in feature_idx
        score_new:      values of score function at x_new = x + a = w0 + w.dot(x + a)
        yhat_new:       value of prediction at x_new = x + a = f(score_new)
        feasible:       True if x + a is a feasible action
        flipped:        True if the f(x) != f(x+a)
        """
        return self._df

    @property
    def yhat(self):
        return self._intercept + np.dot(self._coefs, self._x)

    def predict(self, actions=None):
        return np.sign(self.score(actions=actions))

    def score(self, actions=None):
        if actions is not None:
            return self._intercept + np.dot(self._coefs, self._x + actions)
        else:
            return self._intercept + np.dot(self._coefs, self._x)

    def populate(
        self,
        total_items=10,
        enumeration_type="distinct_subsets",
        cost_type="local",
        time_limit=None,
        node_limit=None,
        display_flag=None,
    ):
        """
        Generates a list of actions to flip the predicted value of the linear classifier
        from feature vector x.

        :param total_items: maximum # of actions to generate
                            set as float('inf') to enumerate all possible actions

        :param enumeration_type: enumeration algorithm to use for the flipset
                            must be a string in Flipset.valid_enumeration_types
                            - 'distinct_subsets'
                            - 'mutually_exclusive'

        :param cost_type: cost function to use for Flipset generation
                          must be a string in Flipset.valid_cost_types
                            options include:
                            - local

        :param time_limit: max # of seconds to spend before stopping the solver at each
                           iteration

        :param node_limit: max # of branch and bound nodes to process before stopping
                           the solver at each iteration

        :param display_flag: True to display solver progress during enumeration

        :return:
        """
        assert enumeration_type in self._valid_enumeration_types, (
            "enumeration_type must be one of %r" % self._valid_enumeration_types
        )

        assert cost_type in self._valid_cost_types, (
            "cost_type must be one of %r" % self._valid_cost_types
        )

        if self._builder is None:
            self._builder = RecourseBuilder(
                action_set=self.action_set,
                x=self.x,
                coefficients=self._coefs,
                intercept=self._intercept,
                mip_cost_type=cost_type,
                solver=self._solver,
            )

        items = self._builder.populate(
            total_items=total_items,
            enumeration_type=enumeration_type,
            time_limit=time_limit,
            node_limit=node_limit,
            display_flag=display_flag,
        )
        self._add(items)
        return self

    def sort(self, **kwargs):
        """
        Reorders the items in the Flipset dataframe
        Arguments used to sort are saved
        :param sort_args: list of fields to use in sort, or arguments passed to
                          pd.DataFrame.sort_values
        :return:
        """
        if len(kwargs) == 0:
            self._df.sort_values(**self._sort_args)
            return

        if "by" in kwargs:
            sort_names = kwargs["by"]
        else:
            sort_names = list(kwargs.keys())

        assert isinstance(sort_names, list)
        assert len(sort_names) > 0
        for s in sort_names:
            assert isinstance(s, str)
            assert s in self._df.columns

        sort_args = {
            "by": sort_names,
            "inplace": kwargs.get("inplace", True),
            "axis": kwargs.get("axis", 0),
        }
        self._df.sort_values(**sort_args)
        self._sort_args = sort_args

    def view(self):
        """
        prints Flipset as Pandas dataframe
        :return:
        """
        return self._df

    def to_flat_df(self):
        """Flatten out the actionsets in the flipset to product either a latex or HTML representation."""
        self.sort()
        tex_columns = ["features", "x", "x_new"]
        tex_df = self._df[tex_columns]
        if len(tex_df) == 0:
            return []

        # split components for each item
        tex_df = tex_df.reset_index().rename(columns={"index": "item_id"})
        df_list = []
        for n in tex_columns:
            tmp = tex_df.set_index(["item_id"])[n].apply(pd.Series).stack()
            tmp = tmp.reset_index().rename(columns={"level_1": "var_id"})
            tmp_name = tmp.columns[-1]
            tmp = tmp.rename(columns={tmp_name: n})
            df_list.append(tmp)

        # combine into a flattened list
        flat_df = df_list[0]
        for k in range(1, len(df_list)):
            flat_df = flat_df.merge(df_list[k])

        # drop the merge index
        flat_df = flat_df.drop(columns=["var_id"])

        # index items by item_id
        flat_df = flat_df.sort_values(by="item_id")
        flat_df = flat_df.rename(
            columns={
                "item_id": "item",
                "features": "Features to Change",
                "x": "Current Value",
                "x_new": "Required Value",
            }
        )
        return flat_df.set_index("item")

    def to_latex(self, name_formatter="\\textit"):
        """
        converts current Flipset to Latex table
        :param name_formatter:
        :return:
        """
        flat_df = self.to_flat_df()

        # add another column for the latex arrow symbol
        idx = flat_df.columns.tolist().index("Required Value")
        flat_df.insert(loc=idx, column="to", value=["longrightarrow"] * len(flat_df))

        # name headers
        flat_df = flat_df.rename(
            columns={
                "features": "\textsc{Feature Subset}",
                "Current Value": "\textsc{Current Values}",
                "Required Value": "\textsc{Required Values}",
            }
        )

        # get raw tex table
        table = flat_df.to_latex(
            multirow=True, index=True, escape=False, na_rep="-", column_format="rlccc"
        )

        # manually wrap names with a formatter function
        if name_formatter is not None:
            for v in self._variable_names:
                table = table.replace(v, "%s{%s}" % (name_formatter, v))

        # add the backslash for the arrow
        table = table.replace("longrightarrow", "$\\longrightarrow$")

        # minor embellishments
        table = table.split("\n")
        table[2] = table[2].replace("to", "")
        table[2] = table[2].replace("{}", "")
        table.pop(3)
        table.pop(3)
        return "\n".join(table)

    def to_html(self):
        # remove the numbering on the left if possible?
        cfpb_color = "#e2efd8"

        def _color_white_or_gray(row):
            first_item = row.name if isinstance(row.name, int) else row.name[0]
            color = "white" if first_item % 2 == 1 else cfpb_color
            res = "background-color: %s" % color
            return [res] * len(row)

        flat_df = self.to_flat_df()
        if len(flat_df) == 0:
            style = (
                "text-shadow: 0px 1px 1px #4d4d4d; color: 'black'; font: 30px 'LeagueGothicRegular'; background-color:"
                + cfpb_color
            )
            ## style 1
            html = (
                pd.DataFrame([{"outcome": "No Recourse"}])
                .style.set_table_styles(
                    [{"selector": "tr", "props": [("background-color", "white")]}]
                )
                .apply(_color_white_or_gray, axis=1)
                .hide_index()
                .render()
            )
            ## style 2
            html = '<span style="' + style + '">No Recourse</span>'
            return html

        # add another column for the latex arrow symbol
        idx = flat_df.columns.tolist().index("Required Value")

        flat_df.insert(loc=idx, column="to", value=["&#8594;"] * len(flat_df))

        idx = (
            pd.DataFrame(flat_df.index)
            .assign(row=lambda df: df.groupby("item").cumcount().pipe(lambda s: s + 1))
            .pipe(lambda df: list(zip(df["item"], df["row"])))
        )

        idx = pd.MultiIndex.from_tuples(idx)
        flat_df.index = idx
        flat_df["Current Value"] = flat_df["Current Value"].apply(
            lambda x: str(int(x)) if int(x) == x else str(x)
        )
        flat_df["Required Value"] = flat_df["Required Value"].apply(
            lambda x: str(int(x)) if int(x) == x else str(x)
        )
        html = (
            flat_df.style.set_table_styles(
                [{"selector": "tr", "props": [("background-color", "white")]}]
            )
            .apply(_color_white_or_gray, axis=1)
            .hide_index()
            .render()
        )
        return html

    def _add(self, items):
        """
        :param items: adds new items to flipset
        :return:
        """
        if isinstance(items, dict):
            items = [items]
        assert isinstance(items, list)
        items = list(map(lambda i: self._validate_item(i), items))
        self._items.extend(items)
        self._add_to_df(items)

    def _validate_item(self, item):
        """
        checks item to be added to the current Flipset
        :param item: raw flipset item
        :return: item in correct format
        """
        assert isinstance(item, dict)
        required_fields = ["feasible", "actions", "cost"]
        for k in required_fields:
            assert k in item, "item missing field %s" % k
        item["actions"] = self._validate_action(item["actions"])
        assert item["cost"] > 0.0, "total cost must be positive"
        assert item["feasible"], "item must be feasible"
        return item

    def _validate_action(self, a):
        """
        checks action vector to the added to the current Flipset
        :param a: action vector
        :return: a or AssertionError
        """
        a = np.array(a, dtype=np.float_).flatten()
        assert len(a) == self._n_variables, (
            "action vector must have %d elements" % self.n_variables
        )
        assert np.isfinite(a).all(), "actions must be finite"
        assert np.count_nonzero(a) >= 1, "at least one action element must be non zero"
        assert np.not_equal(self.yhat, self.predict(a)), (
            "actions do not flip the prediction from %d" % self.yhat
        )
        return a

    def _add_to_df(self, items):
        if len(items) > 0:
            row_data = list(map(lambda item: self._item_to_df_row(item), items))

            self._df = pd.concat(
                [self._df, pd.Series(row_data[0])], ignore_index=True, sort=True
            )[self._df.columns.tolist()]

            # self._df = self._df.append(row_data, ignore_index=True, sort=True)[
            #     self._df.columns.tolist()
            # ]
            self.sort()

    def _item_to_df_row(self, item):
        """
        converts item to a row in the data frame
        :param item:
        :return:
        """
        x = self.x
        a = item["actions"]
        h = self.predict(a)
        nnz_idx = np.flatnonzero(a)
        row = {
            "cost": float(item["cost"]),
            "size": len(nnz_idx),
            "features": [self._variable_names[j] for j in nnz_idx],
            "feature_idx": nnz_idx,
            "x": x[nnz_idx],
            "x_new": x[nnz_idx] + a[nnz_idx],
            "score_new": self.score(a),
            "yhat_new": h,
            "feasible": item["feasible"],
            "flipped": np.not_equal(h, self.yhat),
        }

        return row



from abc import ABC, abstractmethod
from copy import deepcopy
import numpy as np


class BaseRecourse(ABC):
    """
    Base class to define recourse methods.
    """

    def __init__(self, model, threshold=0.5):
        self.model = model
        self.threshold = threshold

    def _get_coefficients(self):
        """Utility function to retrieve model parameters."""

        model = deepcopy(self.model)
        intercept = self.model.intercept_
        coefficients = self.model.coef_

        # Adjusting the intercept to match the desired threshold.
        intercept = intercept - np.log(self.threshold / (1 - self.threshold))
        model.intercept_ = intercept

        return intercept, coefficients, model

    @abstractmethod
    def _counterfactual(self, agent, action_set):
        pass

    def counterfactual(self, population):
        action_set = population.action_set_

        counterfactual_examples = population.data.apply(
            lambda agent: self._counterfactual(agent, action_set), axis=1
        )

        return counterfactual_examples



"""
Implementation of the framework proposed in the paper ``Multi Agent Dynamic
Counterfactual Recourse''.
"""
import sys

try:
    # This variable is injected in the __builtins__ by the build
    # process. It is used to enable importing subpackages of sklearn when
    # the binaries are not built
    # mypy error: Cannot determine type of '__SKLEARN_SETUP__'
    __RECGAME_SETUP__  # type: ignore
except NameError:
    __RECGAME_SETUP__ = False

if __RECGAME_SETUP__:
    sys.stderr.write("Partial import of recgame during the build process.\n")
    # We are not importing the rest of recourse-game during the build
    # process, as it may not be compiled yet
else:
    from . import recourse, populations, environments
    from ._version import __version__

    __all__ = ["recourse", "populations", "environments", "utils", "__version__"]



"""
Adapted from:
https://github.com/ustunb/actionable-recourse/blob/master/recourse/builder.py
"""

import time
import warnings
from itertools import chain
from collections import defaultdict
import numpy as np

from .utils import parse_classifier_args
from ...populations.action_set import ActionSet
from .cplex_helper import (
    Cplex,
    SparsePair,
    set_cpx_parameters,
    set_cpx_display_options,
    set_cpx_time_limit,
    set_cpx_node_limit,
    DEFAULT_CPLEX_PARAMETERS,
)


# todo fix bug when all points are non-zero but non-action
# (demo_credit_script shouldn't run)
class RecourseBuilder(object):
    _default_check_flag = True
    _default_print_flag = False  # True
    _default_node_limit = float("inf")
    _default_time_limit = float("inf")
    _default_mip_cost_type = "max"
    _default_enumeration_type = "distinct_subsets"
    _valid_mip_cost_types = {"total", "local", "max"}
    _valid_enumeration_types = {"mutually_exclusive", "distinct_subsets"}

    def __new__(cls, **kwargs):
        """Factory Method."""
        return super().__new__(_RecourseBuilder)

    def __init__(self, action_set, x=None, **kwargs):
        """
        :param x: vector of input variables for person x
        :param intercept: intercept value of score function
        :param coefs: coefficients of score function
        :param action_set: action set
        :param params: parameters for flipset form/flipset generation
                       (e.g. type of cost function to use / max items etc.)
        """

        # attach coefficients
        self._coefficients, self._intercept = parse_classifier_args(**kwargs)

        # attach action set
        assert isinstance(action_set, ActionSet)
        assert len(action_set) == len(self._coefficients)
        if not action_set.alignment_known:
            action_set.set_alignment(self._coefficients)

        self._action_set = action_set

        # add indices
        self._variable_names = action_set.name
        self._variable_index = {n: j for j, n in enumerate(self._variable_names)}
        self._actionable_indices = [j for j, v in enumerate(action_set.actionable) if v]

        # flags
        self.print_flag = kwargs.get("print_flag", self._default_print_flag)
        self.check_flag = kwargs.get("check_flag", self._default_check_flag)

        # initialize Cplex MIP
        self._mip = None
        self._mip_cost_type = kwargs.get("mip_cost_type", self._default_mip_cost_type)
        self._min_items = 0
        self._max_items = self.n_actionable

        self._apriori_infeasible = (
            False  ## useful if we want to hardcode conditions where
        )
        ## we won't run the MIP, like empty actionsets.
        # attach features
        self._x = None
        if x is not None:
            self.x = x

        assert self._check_rep()

    def _check_rep(self):
        """
        :return: True if representation invariants are true
        """
        if self.check_flag:
            assert self.n_variables == len(self._coefficients)
            assert self.n_variables == len(self._action_set)
            assert self._x is None or self.n_variables == len(self._x)
            assert isinstance(self._intercept, float)
            assert self.action_set.alignment_known
            assert 0 <= self._min_items <= self._max_items <= self.n_variables
        return True

    @property
    def print_flag(self):
        return self._print_flag

    @print_flag.setter
    def print_flag(self, flag):
        if flag is None:
            self._print_flag = bool(self._default_print_flag)
        elif isinstance(flag, bool):
            self._print_flag = bool(flag)
        else:
            raise AttributeError("print_flag must be boolean or None")

    @property
    def check_flag(self):
        return self._check_flag

    @check_flag.setter
    def check_flag(self, flag):
        if flag is None:
            self._check_flag = bool(self._default_check_flag)
        elif isinstance(flag, bool):
            self._check_flag = bool(flag)
        else:
            raise AttributeError("check_flag must be boolean or None")

    def __repr__(self):
        return str(self.action_set)

    #### immutable properties ####
    @property
    def action_set(self):
        return self._action_set

    @property
    def variable_names(self):
        return self._variable_names

    @property
    def n_variables(self):
        return len(self._coefficients)

    @property
    def variable_index(self):
        return self._variable_index

    @property
    def actionable_indices(self):
        return self._actionable_indices

    @property
    def n_actionable(self):
        return len(self._actionable_indices)

    #### feature values ####
    @property
    def x(self):
        return np.array(self._x)

    @x.setter
    def x(self, x):
        assert isinstance(x, (list, np.ndarray))
        x = np.array(x, dtype=np.float_).flatten()
        assert len(x) == self.n_variables
        self._x = x
        self.build_mip()

    #### model ####
    @property
    def coefficients(self):
        return self._coefficients

    @property
    def intercept(self):
        return float(self._intercept)

    def score(self, x=None):
        if x is None:
            x = self._x
        else:
            assert isinstance(x, (np.ndarray, list))
            x = np.array(x, dtype=np.float).flatten()
        return self._coefficients.dot(x) + self._intercept

    def prediction(self, x=None):
        return np.sign(self.score(x))

    #### MIP ####
    @property
    def mip(self):
        if self._mip is None:
            raise ValueError("mip has not yet been initialized")
        return self._mip

    @property
    def mip_cost_type(self):
        return self._mip_cost_type

    @property
    def mip_display(self):
        return self._mip_display_flag

    @property
    def mip_time_limit(self):
        return self._mip_time_limit

    @property
    def mip_node_limit(self):
        return self._mip_node_limit

    @mip_time_limit.setter
    def mip_time_limit(self, value):
        if value is None:
            value = self._default_time_limit
        elif isinstance(value, (float, int)) and value > 0.0:
            value = float(value)
        else:
            raise ValueError("time limit must be positive or inf")
        self._set_mip_time_limit(self.mip, value)
        self._mip_time_limit = value

    @mip_node_limit.setter
    def mip_node_limit(self, value):
        if value is None:
            value = self._default_node_limit
        elif isinstance(value, (float, int)) and value > 0:
            value = float(value)
        else:
            raise ValueError("node limit must be positive integer or inf")
        self._set_mip_node_limit(self.mip, value)
        self._mip_node_limit = value

    @mip_display.setter
    def mip_display(self, flag):
        """
        sets display flag for mip
        :param display_flag:
        :return:
        """
        if flag is None:
            flag = self._default_print_flag
        elif not isinstance(flag, bool):
            raise ValueError("mip_display must be boolean")
        self._set_mip_display(self.mip, flag)
        self._mip_node_limit = flag

    #### mip item limits ####
    def set_mip_item_limits(self, min_items=None, max_items=None):
        """
        changes limits for the number of items
        :param min_items:
        :param max_items:
        :return:
        """

        if min_items is None or max_items is None:
            return

        min_items = self.min_items if min_items is None else int(min_items)
        max_items = self.max_items if min_items is None else int(max_items)
        assert min_items <= max_items, "incompatible sizes"

        min_items = int(min_items)
        if min_items != self.min_items:
            min_nnz_actions = float(self.n_actionable - min_items)
            self.set_mip_min_items(min_nnz_actions)
            self.min_items = min_items

        if max_items != self.max_items:
            max_nnz_actions = float(self.n_actionable - max_items)
            self.set_mip_max_items(min_nnz_actions)
            self.max_items = max_items

        return

    @property
    def min_items(self):
        return int(self._min_items)

    @min_items.setter
    def min_items(self, k):
        if k is None:
            self._min_items = 0
        else:
            k = int(k)
            assert k >= 0
            self._min_items = k

    @property
    def max_items(self):
        return int(self._max_items)

    @max_items.setter
    def max_items(self, k):
        if k is None:
            self._max_items = self.n_actionable
        else:
            k = int(k)
            assert k >= 0
            self._max_items = min(k, self.n_actionable)

    #### mip stamping ####
    def _get_mip_build_info(self, cost_function_type="percentile", validate=True):
        assert cost_function_type == "percentile"

        build_info = {}
        indices = defaultdict(list)

        # get names of constrained variables
        constrained_names = self.action_set.constraints.constrained_names()

        # setup cost function
        if self.mip_cost_type == "local":
            cost_up = lambda c: np.log((1.0 - c[0]) / (1.0 - c))
            cost_dn = lambda c: np.log((1.0 - c) / (1.0 - c[0]))
        else:
            cost_up = lambda c: c - c[0]
            cost_dn = lambda c: c[0] - c

        # todo: set this to returns_compatible = True and check if that changes anything.
        actions, percentiles = self._action_set.feasible_grid(
            x=self._x,
            return_actions=True,
            return_percentiles=True,
            return_compatible=True,
        )

        for n, a in actions.items():
            if len(a) >= 2:
                c = percentiles[n]
                if np.isclose(a[-1], 0.0):
                    a = np.flip(a, axis=0)
                    c = np.flip(c, axis=0)
                    c = cost_dn(c)
                else:
                    c = cost_up(c)

                # override numerical issues
                bug_idx = np.logical_or(
                    np.less_equal(c, 0.0), np.isclose(a, 0.0, atol=1e-8)
                )
                bug_idx = np.flatnonzero(bug_idx).tolist()
                bug_idx.pop(0)
                if len(bug_idx) > 0:
                    c = np.delete(c, bug_idx)
                    a = np.delete(a, bug_idx)

                idx = self._variable_index[n]
                w = float(self._coefficients[idx])
                # da = np.diff(a)
                dc = np.diff(c)

                # find nullifying index
                # k_null is index such that x[j] - a[j][k_null] = 0
                # action at nullifying index sets the variable to 0
                x = self._x[idx]
                if n in constrained_names and x > 0:
                    k_null = np.flatnonzero(a == -x)[0]
                else:
                    k_null = 0

                info = {
                    "idx": idx,
                    "coef": w,
                    "actions": a.tolist(),
                    "costs": c.tolist(),
                    "action_var_name": ["a[%d]" % idx],
                    "action_ind_names": ["u[%d][%d]" % (idx, k) for k in range(len(a))],
                    "cost_var_name": ["c[%d]" % idx],
                    "nullify_ind_name": ["u[%d][%d]" % (idx, k_null)],
                }

                build_info[n] = info

                indices["var_idx"].append(idx)
                indices["coefficients"].append(w)
                indices["action_off_names"].append(
                    info["action_ind_names"][0]
                )  ## the indices of variables that indicate that the feature is "off", i.e. no actions are taken on that feature.
                indices["action_ind_names"].extend(info["action_ind_names"])
                #
                indices["nullify_ind_names"].append(info["nullify_ind_name"][0])
                # indices['nullify_ind_values'].append(k_null)
                #
                indices["action_var_names"].extend(info["action_var_name"])
                indices["cost_var_names"].extend(info["cost_var_name"])
                indices["action_lb"].append(float(np.min(a)))
                indices["action_ub"].append(float(np.max(a)))
                # indices['action_df'].append(float(np.min(da)))
                indices["cost_ub"].append(float(np.max(c)))
                indices["cost_df"].append(float(np.min(dc)))

        # get names of variables associated with constraints
        if validate:
            assert self._check_mip_build_info(build_info)

        return build_info, indices

    def _check_mip_build_info(self, build_info):
        y_desired = self.action_set.y_desired
        constrained_names = self.action_set.constraints.constrained_names()
        for k, v in build_info.items():
            assert not np.isclose(v["coef"], 0.0)
            a = np.array(v["actions"])
            c = np.array(v["costs"])
            assert c[0] == 0.0
            assert a[0] == 0.0

            if k not in constrained_names:
                if y_desired * np.sign(v["coef"]) > 0:
                    assert np.all(np.greater(a[1:], 0.0))
                else:
                    assert np.all(np.less(a[1:], 0.0))

            assert len(a) >= 2
            assert len(a) == len(c)
            assert len(a) == len(np.unique(a))
            assert len(c) == len(np.unique(c))

        return True

    #### mip solutions ####
    @property
    def _empty_mip_solution_info(self):
        """
        :return: dictionary containing the output of the MIP.
                 default values correspond to a case where the model is infeasible
        """

        info = {
            "cost": float("inf"),
            "feasible": False,
            "status": "no solution exists",
            #
            "costs": np.repeat(np.nan, self.n_variables),
            "actions": np.repeat(np.nan, self.n_variables),
            "upperbound": float("inf"),
            "lowerbound": float("inf"),
            "gap": float("inf"),
            #
            "iterations": 0,
            "nodes_processed": 0,
            "nodes_remaining": 0,
            "runtime": 0.0,
        }

        return info

    def _check_mip_solution(self, info):
        """
        :return: return True if making the change from the Flipset will actually 'flip' the prediction for the classifier
        """
        if info["feasible"]:
            a = info["actions"]
            all_idx = np.arange(len(a))
            static_idx = np.flatnonzero(np.isclose(a, 0.0, rtol=1e-4))
            action_idx = np.setdiff1d(all_idx, static_idx)
            n_items = len(action_idx)
            assert n_items >= 1
            assert self.min_items <= n_items <= self.max_items

            try:
                assert np.all(np.isin(action_idx, self.actionable_indices))
            except AssertionError:
                warnings.warn("action set no in self.actionable_indices")

            x = self.x
            try:
                assert np.not_equal(self.prediction(x), self.prediction(x + a))
            except AssertionError:
                s = self.score(x + a)
                assert not np.isclose(self.score(x + a), 0.0, atol=1e-4)
                warnings.warn("numerical issue: near-zero score(x + a) = %1.8f" % s)

            try:
                # check costs change -> action
                assert np.all(np.greater(info["costs"][action_idx], 0.0))
                assert np.all(np.isclose(info["costs"][static_idx], 0.0))

                # check total cost
                if self.mip_cost_type == "max":
                    if not np.isclose(info["cost"], np.max(info["costs"]), rtol=1e-4):
                        warnings.warn(
                            "numerical issue: max_cost is %1.2f but maximum of cost[j] is %1.2f"
                            % (info["cost"], np.max(info["costs"]))
                        )
                elif self.mip_cost_type == "total":
                    assert np.isclose(info["cost"], np.sum(info["costs"]))

            except AssertionError:
                warnings.warn("issue detected with %s" % str(info))

        return True

    def fit(self, time_limit=None, node_limit=None, display_flag=False):
        """
        Solve recourse problem
        :param time_limit:
        :param node_limit:
        :param display_flag:
        :return:
        """
        if self._apriori_infeasible:
            return self._empty_mip_solution_info

        assert self._mip is not None, "must first initialize recourse IP"

        # update time limit and node limit
        self.mip_display = display_flag
        self.mip_time_limit = time_limit
        self.mip_node_limit = node_limit

        # solve
        start_time = time.process_time()
        self.solve_mip()
        end_time = time.process_time() - start_time
        info = self.solution_info
        info["runtime"] = end_time
        return info

    def populate(
        self,
        total_items=10,
        enumeration_type="distinct_subsets",
        time_limit=None,
        node_limit=None,
        display_flag=False,
    ):
        """
        Repeatedly solve recourse problem to recover successive optima.

            :param total_items: int. The number of different actionsets to return.

            :param enumeration_type str: either {"mutually_exclusive", "distinct_subsets"}.
                Specifies how to generate different actions to suggest to flip a prediction.
                  * "mutually_exclusive": each actionset contains wholly different features than those previously used.
                        ex) actionset_i is: change features [1, 4, 8].
                            actionset_{j!=i} does not contain actions on features 1, 4 or 8.
                  * "distinct_subsets": no actionset repeats the _same_ feature combination.
                        ex) actionset_i is: change features [1, 4, 8].
                            actionset_{j!=i} does not contain actions on features 1, 4, and 8.

            :param time_limit: the time limit on the solver.

            :param node_limit: the node limit on the solver.

            :param display_flag: the verbosity of the solver.

            :return: flipset: A set of actionsets.
        """
        assert self._mip is not None, "must first initialize recourse IP"
        assert (isinstance(total_items, int) and total_items >= 1) or (
            isinstance(total_items, float) and total_items == float("inf")
        )
        assert enumeration_type in self._valid_enumeration_types

        # set the remove solution method
        if enumeration_type == "mutually_exclusive":
            remove_solution = self.remove_all_features
        else:
            remove_solution = self.remove_feature_combination

        # update time limit and node limit
        self.mip_display = display_flag
        self.mip_time_limit = time_limit
        self.mip_node_limit = node_limit

        # enumerate soluitions
        k = 0
        all_info = []
        populate_start_time = time.process_time()
        while k < total_items:
            self.solve_mip()
            info = self.solution_info
            if not info["feasible"]:
                if self.print_flag:
                    print("recovered all minimum-cost items")
                break
            all_info.append(info)
            remove_solution()
            k += 1

        if self.print_flag:
            print(
                "obtained %d items in %1.1f seconds"
                % (k, time.process_time() - populate_start_time)
            )

        return all_info


class _RecourseBuilder(RecourseBuilder):
    _default_cplex_parameters = dict(DEFAULT_CPLEX_PARAMETERS)

    def __init__(self, action_set, x=None, **kwargs):
        """
        :param x: vector of input variables for person x
        :param intercept: intercept value of score function
        :param coefs: coefficients of score function
        :param action_set: action set
        :param params: parameters for flipset form/flipset generation
                       (e.g. type of cost function to use / max items etc.)
        """
        # initialize Cplex MIP
        self._cpx_parameters = kwargs.get(
            "cplex_parameters", self._default_cplex_parameters
        )
        self._set_mip_time_limit = set_cpx_time_limit
        self._set_mip_node_limit = set_cpx_node_limit
        self._set_mip_display = lambda mip, flag: set_cpx_display_options(
            mip, display_lp=flag, display_mip=flag, display_parameters=flag
        )

        ## initialize base class
        super().__init__(action_set=action_set, x=x, **kwargs)

    #### building MIP ####
    def build_mip(self):
        """
        returns an optimization problem that can be solved to determine an item in a flipset for x
        :return:
        """

        # MIP parameters
        cost_type = self.mip_cost_type
        min_items = max(self.min_items, 1)
        max_items = self.max_items

        # cost/action information
        build_info, indices = self._get_mip_build_info()
        ## TODO: note: if actiongrid is empty, build_info, indices == {}. Correct handling?

        # initialize mip
        mip = Cplex()
        mip.set_problem_type(mip.problem_type.MILP)
        vars = mip.variables
        cons = mip.linear_constraints
        n_actionable = len(build_info)
        n_indicators = len(indices["action_ind_names"])

        # define a[j]
        vars.add(
            names=indices["action_var_names"],
            types=["C"] * n_actionable,
            lb=indices["action_lb"],
            ub=indices["action_ub"],
        )

        # score constraint
        # y_desired = +1:  sum_j w[j]*(x[j]+a[j]) > 0 -> sum_j w[j] a[j] > -score
        # y_desired = -1:  sum_j w[j]*(x[j]+a[j]) < 0 -> sum_j w[j] a[j] < -score
        score_constraint_sense = "G" if self.action_set.y_desired > 0 else "L"
        cons.add(
            names=["score"],
            lin_expr=[
                SparsePair(ind=indices["action_var_names"], val=indices["coefficients"])
            ],
            senses=[score_constraint_sense],
            rhs=[-self.score()],
        )

        # define indicators u[j][k] = 1 if a[j] = actions[j][k]
        vars.add(names=indices["action_ind_names"], types=["B"] * n_indicators)

        # restrict a[j] to feasible values using a 1 of K constraint setup
        # restrict a[j] to actions in feasible set and make sure exactly 1 indicator u[j][k] is on
        for info in build_info.values():
            # 1. a[j]  =   sum_k u[j][k] * actions[j][k] -> 0.0   =   sum u[j][k] * actions[j][k] - a[j]
            # 2. sum_k u[j][k] = 1.0
            cons.add(
                names=["set_a[%d]" % info["idx"], "pick_a[%d]" % info["idx"]],
                lin_expr=[
                    SparsePair(
                        ind=info["action_var_name"] + info["action_ind_names"],
                        val=[-1.0] + info["actions"],
                    ),
                    SparsePair(
                        ind=info["action_ind_names"], val=[1.0] * len(info["actions"])
                    ),
                ],
                senses=["E", "E"],
                rhs=[0.0, 1.0],
            )

            # declare indicator variables as SOS set
            mip.SOS.add(
                type="1",
                name="sos_u[%d]" % info["idx"],
                SOS=SparsePair(ind=info["action_ind_names"], val=info["actions"]),
            )

        # limit number of features per action
        #
        # size := n_actionable - n_null where n_null := sum_j u[j][0] = sum_j 1[a[j] = 0]
        #
        # size <= max_size
        # n_actionable - sum_j u[j][0]  <=  max_size
        # n_actionable - max_size       <=  sum_j u[j][0]
        #
        # min_size <= size:
        # min_size          <=  n_actionable - sum_j u[j][0]
        # sum_j u[j][0]     <=  n_actionable - min_size
        size_expr = SparsePair(
            ind=indices["action_off_names"], val=[1.0] * n_actionable
        )
        cons.add(
            names=["max_items", "min_items"],
            lin_expr=[size_expr, size_expr],
            senses=["G", "L"],
            rhs=[float(n_actionable - max_items), float(n_actionable - min_items)],
        )

        for c in self.action_set.constraints:
            """
            the constraint has the form:

            c.lb - k <= -\sum_ {j \ in indices} u[j][k_null] <= c.ub - k

            where, k_null is defined as the value such that:

            x[j] + a[j][k_null] = 0

            the constraint is implemeneted in CPLEX using a ranged constraint of the form:

            lhs[i] <= a*x[i] <= lhs[i] + range_values[i]

            here:

            lhs = float(c.lb - k)
            rhs = float(c.ub - k)
            range_values = rhs - lhs = float(c.ub - c.lb)
            """
            k = len(c.indices)
            null_names = [indices["nullify_ind_names"][j] for j in c.indices]
            cons.add(
                names=["subset_limit_%s" % c.id],
                lin_expr=[SparsePair(ind=null_names, val=[-1.0] * k)],
                senses=["R"],
                rhs=[float(c.lb) - k],
                range_values=[float(c.ub - c.lb)],
            )

        # add constraints for cost function
        if cost_type in ("total", "local"):
            indices.pop("cost_var_names")
            objval_pairs = list(
                chain(
                    *[
                        list(zip(v["action_ind_names"], v["costs"]))
                        for v in build_info.values()
                    ]
                )
            )
            mip.objective.set_linear(objval_pairs)

        elif cost_type == "max":
            indices["max_cost_var_name"] = ["max_cost"]

            ## handle empty actionsets
            indices["epsilon"] = np.min(indices["cost_df"] or np.inf) / np.sum(
                indices["cost_ub"]
            )
            vars.add(
                names=indices["max_cost_var_name"] + indices["cost_var_names"],
                types=["C"] * (n_actionable + 1),
                obj=[1.0] + [indices["epsilon"]] * n_actionable,
            )

            # lb = [0.0] * (n_actionable + 1)) # default values are 0.0
            cost_constraints = {
                "names": [],
                "lin_expr": [],
                "senses": ["E", "G"] * n_actionable,
                "rhs": [0.0, 0.0] * n_actionable,
            }

            for info in build_info.values():
                cost_constraints["names"].extend(
                    ["def_cost[%d]" % info["idx"], "set_max_cost[%d]" % info["idx"]]
                )

                cost_constraints["lin_expr"].extend(
                    [
                        SparsePair(
                            ind=info["cost_var_name"] + info["action_ind_names"],
                            val=[-1.0] + info["costs"],
                        ),
                        SparsePair(
                            ind=indices["max_cost_var_name"] + info["cost_var_name"],
                            val=[1.0, -1.0],
                        ),
                    ]
                )

            cons.add(**cost_constraints)

            # old code (commented out for speed)
            #
            # vars.add(names = indices['cost_var_names'],
            #          types = ['C'] * n_actionable,
            #          obj = [indices['epsilon']] * n_actionable,
            #          #ub = [CPX_INFINITY] * n_actionable, #indices['cost_ub'], #indices['cost_ub'],
            #          lb = [0.0] * n_actionable)
            #
            # vars.add(names = indices['max_cost_var_name'],
            #          types = ['C'],
            #          obj = [1.0],
            #          #ub = [np.max(indices['cost_ub'])],
            #          lb = [0.0])
            #
            # for info in build_info.values():
            #     cost[j] = sum c[j][k] u[j][k]
            #     cons.add(names = ['def_cost[%d]' % info['idx']],
            #              lin_expr = [SparsePair(ind = info['cost_var_name'] + info['action_ind_names'], val = [-1.0] + info['costs'])]
            #              senses = ["E"],
            #              rhs = [0.0])
            #
            #     max_cost > cost[j]
            #     cons.add(names = ['set_max_cost[%d]' % info['idx']],
            #              lin_expr = [SparsePair(ind = indices['max_cost_var_name'] + info['cost_var_name'], val = [1.0, -1.0])],
            #              senses = ["G"],
            #              rhs = [0.0])

        mip = set_cpx_parameters(mip, self._cpx_parameters)
        self._mip = mip
        self._mip_indices = indices

    #### MIP settings ###
    def set_mip_parameters(self, param=None):
        """
        updates MIP parameters
        :param param:
        :return:
        """
        if param is None:
            param = self._cpx_parameters

        self._mip = set_cpx_parameters(self._mip, param)

    #### solving MIP ###
    def solve_mip(self):
        self.mip.solve()

    @property
    def solution_info(self):
        assert hasattr(self._mip, "solution")
        mip = self._mip
        sol = mip.solution
        info = self._empty_mip_solution_info
        if sol.is_primal_feasible():
            indices = self._mip_indices
            variable_idx = indices["var_idx"]

            # parse actions
            action_values = sol.get_values(indices["action_var_names"])

            if "cost_var_names" in indices and self.mip_cost_type != "total":
                cost_values = sol.get_values(indices["cost_var_names"])
            else:
                ind_idx = np.flatnonzero(
                    np.array(sol.get_values(indices["action_ind_names"]))
                )
                ind_names = [indices["action_ind_names"][int(k)] for k in ind_idx]
                cost_values = mip.objective.get_linear(ind_names)

            actions = np.zeros(self.n_variables)
            np.put(actions, variable_idx, action_values)

            costs = np.zeros(self.n_variables)
            np.put(costs, variable_idx, cost_values)

            info.update(
                {
                    "feasible": True,
                    "status": sol.get_status_string(),
                    #
                    "actions": actions,
                    "costs": costs,
                    #
                    "upperbound": sol.get_objective_value(),
                    "lowerbound": sol.MIP.get_best_objective(),
                    "gap": sol.MIP.get_mip_relative_gap(),
                    #
                    "iterations": sol.progress.get_num_iterations(),
                    "nodes_processed": sol.progress.get_num_nodes_processed(),
                    "nodes_remaining": sol.progress.get_num_nodes_remaining(),
                }
            )

            if self.mip_cost_type == "max":
                info["cost"] = sol.get_values(indices["max_cost_var_name"])[0]
            else:
                info["cost"] = info["upperbound"]

        else:
            info.update(
                {
                    "iterations": sol.progress.get_num_iterations(),
                    "nodes_processed": sol.progress.get_num_nodes_processed(),
                    "nodes_remaining": sol.progress.get_num_nodes_remaining(),
                }
            )

        return info

    #### flipset geneation ###
    def set_mip_min_items(self, n_items):
        """
        sets maximum number of non-zero elements in MIP (used by set_item_limits)
        :param n_items:
        :return:
        """
        self._mip.linear_constraints.set_rhs("min_items", n_items)

    def set_mip_max_items(self, n_items):
        """
        sets maximum number of non-zero elements in MIP (used by set_item_limits)
        :param n_items:
        :return:
        """
        self._mip.linear_constraints.set_rhs("max_items", n_items)

    def remove_all_features(self):
        """
        removes feature combination from feasible region of MIP
        :return:
        """
        mip = self._mip
        ## "action_off_names" ex: ['u[3][0]', 'u[4][0]'...] are variables that indicate an action is "off".
        feature_off_idxs = self._mip_indices["action_off_names"]
        ## get the values assigned by the solver.
        values_of_off_indices = np.array(mip.solution.get_values(feature_off_idxs))
        ## if the "off index" are off (i.e. = 0), that means the action is "on"
        on_idx = np.flatnonzero(np.isclose(values_of_off_indices, 0.0))
        ## setting LB = 1 for the "off index" means that the action has to stay "off"
        mip.variables.set_lower_bounds([(feature_off_idxs[j], 1.0) for j in on_idx])
        return

    def remove_feature_combination(self):
        """
        removes feature combination from feasible region of MIP
        :return:
        """

        mip = self._mip
        # feature_off_idxs = self._mip_indices['action_off_names']

        # todo: revisit this constraint
        """
        get a = np.array(mip.solution.get_values(action_var_names))
        for j, aj in enumerate(a):
            if aj != 0:
                if self._names[j] is a one-hot-encoding:
                    off_idx[j] = nullify_ind_names[j]
                otherwise:
                    off_idx[j] = action_off_names[j]
        """

        feature_off_idxs = self._mip_indices["nullify_ind_names"]
        u = np.array(mip.solution.get_values(feature_off_idxs))

        ## if the "off index" are off (i.e. = 0), that means the action is "on"
        on_idx = np.isclose(u, 0.0)

        ## array where con_val[i] = 1 if feature is off, -1 if feature is on.
        con_vals = np.ones(len(feature_off_idxs), dtype=np.float_)
        con_vals[on_idx] = -1.0

        ## one minus number of features that are off.
        con_rhs = np.sum(~on_idx) - 1
        mip.linear_constraints.add(
            lin_expr=[SparsePair(ind=feature_off_idxs, val=con_vals.tolist())],
            senses=["L"],
            rhs=[float(con_rhs)],
        )
        return



from .plot_environment import EnvironmentPlot
from .simulations import simulations_scores_panel, simulations_success_rate_panels

__all__ = [
    "EnvironmentPlot",
    "simulations_scores_panel",
    "simulations_success_rate_panels",
]



import numpy as np


def is_sklearn_linear_classifier(obj):
    """
    Checks if object is a sklearn linear classifier for a binary outcome
    :param obj: object
    """
    binary_flag = hasattr(obj, "classes_") and len(obj.classes_) == 2
    linear_flag = hasattr(obj, "coef_") and hasattr(obj, "intercept_")
    return binary_flag and linear_flag


def parse_classifier_args(*args, **kwargs):
    """
    helper function to parse coefficients and intercept from linear classifier arguments

    *args and **kwargs can contain either:
        - sklearn classifiers with 'coef_' and 'intercept_' fields
          (keyword: 'clf', 'classifier')
        - vector of coefficients (keyword: 'coefficients')
        - intercept: set to 0 by default (keyword: 'intercept')

    returns:
        w - np.array containing coefficients of linear classifier (finite, flattened)
        t - float containing intercept of linear classifier (finite, float)

    raises:
        ValueError if fails to parse classifier arguments

    :return:
    """
    w, t = None, None

    if "clf" in kwargs:
        assert is_sklearn_linear_classifier(kwargs["clf"])
        w = kwargs["clf"].coef_
        t = kwargs["clf"].intercept_

    elif "classifier" in kwargs:
        assert is_sklearn_linear_classifier(kwargs["classifier"])
        w = kwargs["classifier"].coef_
        t = kwargs["classifier"].intercept_

    elif "coefficients" in kwargs:
        w = kwargs.get("coefficients")
        t = kwargs.get("intercept", 0.0)

    elif len(args) == 1:
        if is_sklearn_linear_classifier(args[0]):
            w = args[0].coef_
            t = args[0].intercept_

        elif isinstance(args[0], (list, np.ndarray)):
            w = np.array(args[0]).flatten()
            t = 0.0

    elif len(args) == 2:
        w = args[0]
        t = float(args[1])

    else:
        raise ValueError("failed to match classifier arguments")

    w = np.array(w).flatten()
    t = float(t)
    assert np.isfinite(w).all()
    assert np.isfinite(t)
    return w, t



import numpy as np
from rich.progress import track
import matplotlib.pyplot as plt
from scipy.ndimage import gaussian_filter1d


def simulations_scores_panel(results, adaptation_list, new_agents_list):
    """
    TODO:
    - Documentation
    - Remove adaptation_list and new_agents_list from parameters (they are not necessary)
    """
    fig, axes = plt.subplots(
        len(adaptation_list),
        len(new_agents_list),
        figsize=(10, 10),
        layout="constrained",
        sharey=True,
        sharex=True,
    )

    for params, env in track(results):
        i = adaptation_list.index(params["adaptation"])
        j = new_agents_list.index(params["new_agents"])

        ax = env.plot.agent_scores(title=False, legend=False, ax=axes[i, j])
        ax.set_ylabel("")
        ax.set_xlabel("")
        ax.set_xticks(range(0, 51, 10))
        ax.set_xlim(-1, 51)

    for i in range(len(adaptation_list)):
        axes[i, 0].set_ylabel(f"{adaptation_list[i]}")

    for j in range(len(new_agents_list)):
        axes[-1, j].set_xlabel(f"{new_agents_list[j]}")

    fig.supxlabel(r"Number of New Agents")
    fig.supylabel("Adaptation")

    return fig, axes


def simulations_success_rate_panels(results, adaptation_list, new_agents_list):
    """
    TODO:
    - Documentation
    - Remove adaptation_list and new_agents_list from parameters (they are not necessary)
    """
    fig, axes = plt.subplots(
        len(adaptation_list),
        len(new_agents_list),
        figsize=(10, 10),
        layout="constrained",
        sharey=True,
        sharex=True,
    )

    for params, env in track(results):
        i = adaptation_list.index(params["adaptation"])
        j = new_agents_list.index(params["new_agents"])
        sr = env.success_rate(1, env.step_ + 1)
        sr_smooth = gaussian_filter1d(
            np.where(np.isnan(sr), sr[~np.isnan(sr)].mean(), sr), sigma=5
        )

        ax = axes[i, j]
        ax.plot(range(1, env.step_ + 1), sr, alpha=0.25)
        ax.plot(range(1, env.step_ + 1), sr_smooth, c="#d75c5c")
        ax.set_xticks(range(0, 51, 10))
        ax.set_xlim(-1, 51)

    for i in range(len(adaptation_list)):
        axes[i, 0].set_ylabel(f"{adaptation_list[i]}")

    for j in range(len(new_agents_list)):
        axes[-1, j].set_xlabel(f"{new_agents_list[j]}")

    # fig.suptitle("Success rate")
    fig.supxlabel(r"Number of New Agents")
    fig.supylabel("Adaptation")

    return fig, axes



from typing import Union
from copy import deepcopy
import numpy as np
import pandas as pd
from .base import BaseEnvironment
from ..utils import generate_synthetic_data


def _add_agents(self, n_agents):
    all_cols = self.population.data.columns
    categorical = (
        [] if self.population.categorical is None else self.population.categorical
    )
    continuous = all_cols.drop(categorical)

    X, _, _ = generate_synthetic_data(
        n_agents=n_agents,
        n_continuous=len(continuous),
        n_categorical=len(categorical),
        random_state=self._rng,
    )
    return X


# def _add_agents(self, n_agents):
#     all_cols = self.population.data.columns
#     categorical = (
#         [] if self.population.categorical is None else self.population.categorical
#     )
#     continuous = all_cols.drop(categorical)
#
#     new_agents = pd.DataFrame(
#         self._rng.normal(loc=0.5, scale=1/3, size=(n_agents, len(continuous))),
#         columns=continuous,
#         index=range(self._max_id + 1, self._max_id + n_agents + 1),
#     )
#     for col in categorical:
#         # Use the original data to retrieve the distributions
#         counts = self.population.data.groupby(col).size()
#         new_agents[col] = self._rng.choice(
#             counts.index, size=n_agents, p=counts / counts.sum()
#         ).astype(self.population_.data[col].dtype)
#
#     return new_agents


class ClosedEnvironment(BaseEnvironment):
    """
    Closed, general-purpose environment

    TODO: review parameters.
    """

    def __init__(
        self,
        population,
        recourse,
        threshold: float = 0.5,
        threshold_type: str = "fixed",
        adaptation: Union[float, np.ndarray, pd.Series] = 1.0,
        adaptation_type: str = "stepwise",
        remove_winners: bool = True,
        random_state=None,
    ):
        super().__init__(
            population=population,
            recourse=recourse,
            threshold=threshold,
            threshold_type=threshold_type,
            adaptation=adaptation,
            adaptation_type=adaptation_type,
            growth_rate=0,
            remove_winners=remove_winners,
            random_state=random_state,
        )

    def add_agents(self, n_agents):
        pass


class BankLoanApplication1(BaseEnvironment):
    """
    Bank Loan Application Environment 1:
    - Uniformly generated data across the different features;
    - Dynamic threshold;
    - Binary adaptation;
    - Relative growth rate;
    """

    def __init__(
        self,
        population,
        recourse,
        threshold=0.8,
        adaptation=0.2,
        growth_rate=0.2,
        random_state=None,
    ):
        super().__init__(
            population=population,
            recourse=recourse,
            threshold=threshold,
            threshold_type="dynamic",
            adaptation=adaptation,
            adaptation_type="binary",
            growth_rate=growth_rate,
            growth_rate_type="relative",
            remove_winners=True,
            random_state=random_state,
        )

    def add_agents(self, n_agents):
        return _add_agents(self, n_agents)


class BankLoanApplication2(BaseEnvironment):
    """
    Bank Loan Application Environment 2:
    - Uniformly generated data across the different features;
    - Absolute threshold;
    - Binary_fixed adaptation;
    - Absolute growth rate;

    ``n_loans`` is the number of loans granted per time step
    ``new_agents`` is the number of new applicants entering the environment
    """

    def __init__(
        self,
        population,
        recourse,
        n_loans=10,
        adaptation=20,
        new_agents=10,
        random_state=None,
    ):
        super().__init__(
            population=population,
            recourse=recourse,
            threshold=n_loans,
            threshold_type="absolute",
            adaptation=adaptation,
            adaptation_type="binary_fixed",
            growth_rate=new_agents,
            growth_rate_type="absolute",
            remove_winners=True,
            random_state=random_state,
        )

    def add_agents(self, n_agents):
        return _add_agents(self, n_agents)


class BankLoanApplication3(BaseEnvironment):
    """
    Bank Loan Application Environment 3:
    - Uniformly generated data across the different features;
    - Absolute threshold;
    - Gaussian adaptation;
    - Absolute growth rate;

    ``n_loans`` is the number of loans granted per time step.
    ``new_agents`` is the number of new applicants entering the environment.
    ``adaptation`` adjusts the flexibility for agents to adapt.
    """

    def __init__(
        self,
        population,
        recourse,
        n_loans=10,
        adaptation=0.2,
        new_agents=10,
        random_state=None,
    ):
        super().__init__(
            population=population,
            recourse=recourse,
            threshold=n_loans,
            threshold_type="absolute",
            adaptation=adaptation,
            adaptation_type="gaussian",
            growth_rate=new_agents,
            growth_rate_type="absolute",
            remove_winners=True,
            random_state=random_state,
        )

    def add_agents(self, n_agents):
        return _add_agents(self, n_agents)


class WillingnessEnvironment(BaseEnvironment):
    """
    Adapts the gaussian adaptation to include an additional parameter ``willingness`` to
    determine how easily agents can adapt, instead of the function defined with the
    gaussian case.

    Characteristics of this environment:
    - Uniformly generated data across the different features;
    - Absolute threshold;
    - Willingness-based adaptation;
    - Absolute growth rate;

    ``willingness`` should be a numpy vector with individual willingness values.
    """

    def __init__(
        self,
        population,
        recourse,
        n_loans=10,
        adaptation=0.2,
        willingness=None,
        new_agents=10,
        random_state=None,
    ):
        self.willingness = willingness
        super().__init__(
            population=population,
            recourse=recourse,
            threshold=n_loans,
            threshold_type="absolute",
            adaptation=adaptation,
            adaptation_type="gaussian",
            growth_rate=new_agents,
            growth_rate_type="absolute",
            remove_winners=True,
            random_state=random_state,
        )

    def _check(self):
        """Override to use willingness."""
        if self.willingness is not None and not hasattr(self, "adaptation_"):
            self.adaptation_ = deepcopy(self.willingness)
        super()._check()

    def _update_adaptation(self):
        """
        Defining individualized willingness can be based on observation, or be entirely
        arbitrary. In this setting, it is based on the initial score and it is never
        updated (i.e., arbitrary).
        """
        # This is the global adaptation, it is a float value
        adaptation = (
            self.adaptation[self.step_]
            if type(self.adaptation) == [np.ndarray, list]
            else self.adaptation
        )
        current_adaptation = self.adaptation_ if hasattr(self, "adaptation_") else None

        agents = (
            self._new_agents if hasattr(self, "_new_agents") else self.population_.data
        )

        # Define their willingness to adapt based on initial score
        # scores = self.model_.predict_proba(agents)[:, 1]
        # threshold = self.threshold_

        # # ``x`` is the invidual adaptation (willingness), ``adaptation`` is global
        # # in ``x``, with this configuration, lower is better
        # x = threshold - scores
        # adaptation = adaptation / (20 * np.exp(10 * x))
        # adaptation = pd.Series(adaptation, index=agents.index)

        # Alternative, make willingness to adapt random
        x = self._rng.random()
        adaptation = x * adaptation / 10
        adaptation = pd.Series(adaptation, index=agents.index)

        adaptation = pd.concat([current_adaptation, adaptation])

        return pd.Series(adaptation, index=self.population_.data.index)

    def add_agents(self, n_agents):
        return _add_agents(self, n_agents)



import logging
import numpy as np
from .flipset import Flipset
from ..base import BaseRecourse


class ActionableRecourse(BaseRecourse):
    """
    Should work exactly as defined in the paper if the model is linear.
    Otherwise, an explainability model should be used to retrieve coefficients
    for each observation.

    Getting counterfactual is raising multiple warnings, this must be fixed
    later.

    NOTE: Initially only linear models will be considered

    https://arxiv.org/pdf/1809.06514.pdf
    """

    def __init__(
        self, model, threshold=0.5, flipset_size=100, discretize=False, sample=False
    ):
        self.model = model
        self.threshold = threshold
        self.flipset_size = flipset_size
        self.discretize = discretize
        self.sample = sample

    def _counterfactual(self, agent, action_set):
        factual = agent.values
        intercept, coefficients, model = self._get_coefficients()

        # Default counterfactual value if no action flips the prediction
        target_shape = factual.shape[0]
        empty = np.empty(target_shape)
        empty[:] = np.nan
        counterfactual = agent.copy()
        counterfactual.iloc[:] = empty

        # Align action set to coefficients
        action_set.set_alignment(coefficients=coefficients)

        # Build AR flipset
        fs = Flipset(
            x=factual,
            action_set=action_set,
            coefficients=coefficients,
            intercept=intercept,
        )
        try:
            fs_pop = fs.populate(total_items=self.flipset_size)
        except (ValueError, KeyError):
            logging.warning(
                "Actionable Recourse is not able to produce a counterfactual"
                " explanation for instance {}".format(agent.index)
            )
            logging.warning(factual)
            return counterfactual

        # Get actions to flip predictions
        actions = fs_pop.actions

        for action in actions:
            candidate_cf = agent + action

            # Check if candidate counterfactual really flipps the prediction of
            # ML model
            pred_cf = np.argmax(model.predict_proba(candidate_cf.to_frame().T))
            pred_f = np.argmax(model.predict_proba(agent.to_frame().T))
            if pred_cf != pred_f:
                counterfactual.iloc[:] = candidate_cf.squeeze()
                break

        return counterfactual



from abc import ABC, abstractmethod
from typing import Union
from copy import deepcopy
from itertools import product
from collections import Counter
import numpy as np
import pandas as pd
from scipy.stats import binom, norm

from ..visualization import EnvironmentPlot


class BaseEnvironment(ABC):
    """
    Define the constraints in the environment. The central piece of the
    multi-agent analysis of algorithmic recourse.

    ``random_state`` is only relevant for visualizations or if
    ``adaptation_type in ['binary', gaussian]``. For a closed environment (no new
    agents) define ``growth_rate = 0``.

    Some relevant parameters will include:
    - [x] Population
    - [x] Decision model (not necessary)
    - [x] Threshold definition (should accept fixed or dynamic thresholds)
        * Could be defined as a function?
    - [x] Algorithmic Recourse method
    - [x] Distribution of Agents' percentage of adaptation
    - [x] Define how this percentage changes over time:
        * Increase rate: over time people get more motivated to move towards
          counterfactual
        * Decrease rate: over time people lose motivation to move towards
          counterfactual until they eventually give up
        * Mixed: Some people get more motivation, others lose it.
    - [ ] Add verbosity
    - [ ] A lot of functions are using self.model_.predict_proba. Agents' scores could
          be saved in the metadata and repurposed to save on processing time; save a
          ``self.score_`` variable, save it on metadata and use these where
          ``self.model_.predict_proba`` is being used.

    Parameters:
    - population: ``Population`` object containing the Agents' information

    - recourse: Recourse method that allows the production of counterfactuals

    - adaptation: If adaptation_type is stepwise, it represents the rate of adaptation
        of agents towards their counterfactuals. If adaptation_type is binary, the
        adaptation value determines the likelihood of adaptation.  If adaptation_type is
        binary_fixed, the adaptation value determines the number of adaptations.  If
        adaptation_type is gaussian, the adaptation value determines the flexibility of
        agents to adapt.

    - adaptation_type: can be binary, binary_fixed, stepwise, or gaussian

    - threshold: probability threshold to consider an agent's outcome as
      favorable/unfavorable

    - threshold_type: can be fixed, dynamic or absolute

    - growth_rate_type: can be absolute or relative

    - remove_winners: can be True, False

    Attributes:
    - Current step
    -
    """

    def __init__(
        self,
        population,
        recourse,
        threshold: Union[float, int, list, np.ndarray] = 0.5,
        threshold_type: str = "fixed",
        adaptation: Union[float, int, list, np.ndarray] = 1.0,
        adaptation_type: str = "stepwise",
        growth_rate: Union[float, int, list, np.ndarray] = 1.0,
        growth_rate_type: str = "dynamic",
        remove_winners: bool = True,
        random_state=None,
    ):
        self.population = population
        self.recourse = recourse
        self.threshold = threshold
        self.threshold_type = threshold_type
        self.adaptation = adaptation
        self.adaptation_type = adaptation_type
        self.growth_rate = growth_rate
        self.growth_rate_type = growth_rate_type
        self.remove_winners = remove_winners
        self.random_state = random_state

        self.plot = EnvironmentPlot(self, random_state=random_state)
        self._check()
        self._save_metadata()

    def _check(self):
        # the original parameters must be kept unchanged
        if not hasattr(self, "population_"):
            self.population_ = deepcopy(self.population)

        if not hasattr(self, "step_"):
            self.step_ = 0

        if not hasattr(self, "model_"):
            self.model_ = deepcopy(self.recourse.model)

        if not hasattr(self, "growth_k_"):
            self.growth_k_ = self._update_growth_rate()

        if not hasattr(self, "threshold_"):
            self.threshold_ = self._update_threshold()

        if not hasattr(self, "_rng"):
            self._rng = np.random.default_rng(self.random_state)

        if not hasattr(self, "outcome_"):
            self.outcome_, self.scores_ = self.predict(return_scores=True)

        if not hasattr(self, "adaptation_"):
            self.adaptation_ = self._update_adaptation()

        if not hasattr(self, "_max_id"):
            self._max_id = self.population_.data.index.max()

    def _update_threshold(self):
        threshold = (
            self.threshold[self.step_]
            if type(self.threshold) in [np.ndarray, list]
            else self.threshold
        )

        if self.threshold_type == "dynamic":
            self.threshold_index_ = int(
                np.round(threshold * self.population_.data.shape[0])
            )
            probabilities = self.model_.predict_proba(self.population_.data)[:, 1]
            threshold_ = probabilities[np.argsort(probabilities)][self.threshold_index_]

        elif self.threshold_type == "absolute":
            self.threshold_index_ = self.population_.data.shape[0] - threshold
            if self.threshold_index_ < 0:
                self.threshold_index_ = self.population_.data.shape[0] - 1

            probabilities = self.model_.predict_proba(self.population_.data)[:, 1]
            threshold_ = (
                probabilities[np.argsort(probabilities)][self.threshold_index_]
                if threshold != 0
                else np.nan
            )

        elif self.threshold_type == "fixed":
            threshold_ = threshold

        else:
            raise NotImplementedError()

        return threshold_

    def _update_adaptation(self):
        """
        Overwrite this method to implement custom adaptation functions.

        Must return a pd.Series object (where the index corresponds to the population's
        indices).

        NOTE: for the "binary_fixed" case to work properly, this parameter must ALWAYS
        be the last one to be updated.
        """
        adaptation = (
            self.adaptation[self.step_]
            if type(self.adaptation) == [np.ndarray, list]
            else self.adaptation
        )

        if self.adaptation_type == "gaussian":
            # adaptation are the standard deviations for each agent's gaussian
            scores = self.model_.predict_proba(self.population_.data)[:, 1]
            threshold = self.threshold_
            x = threshold - scores
            adaptation = (adaptation) / (10 * np.exp(5 * x))

        elif self.adaptation_type == "binary":
            adaptation = self._rng.binomial(
                1, adaptation, self.population_.data.shape[0]
            )

        elif self.adaptation_type == "binary_fixed":
            idx = np.where(self.scores_ < self.threshold_)[0]

            # If this is false, we assume all remaining agents adapt
            if idx.shape[0] > int(adaptation):
                idx = self._rng.choice(idx, size=int(adaptation), replace=False)

            adaptation = np.zeros(self.population_.data.shape[0], dtype=int)
            adaptation[idx] = 1

        elif self.adaptation_type == "stepwise":
            # No action needed - elif included for clarification
            pass

        return pd.Series(adaptation, index=self.population_.data.index)

    def _update_growth_rate(self):
        """Must return the number of agents to add."""
        growth_rate = (
            self.growth_rate[self.step_]
            if type(self.growth_rate) == [np.ndarray, list]
            else self.growth_rate
        )

        if self.growth_rate_type == "absolute":
            growth_k = int(np.round(growth_rate))
        elif self.growth_rate_type == "relative":
            growth_k = int(np.round(self.population_.data.shape[0] * growth_rate))
        return growth_k

    def predict(self, population=None, step=None, return_scores=False):
        """
        Produces the outcomes for a single agent or a population.

        If ``return_scores=True``, the function will return both the outcome and scores.

        TODO: Refactor if statement
        """

        if step is None:
            threshold = self.threshold_
            if self.threshold_type in ["dynamic", "absolute"]:
                threshold_idx = self.threshold_index_
            if population is None:
                population = self.population_
        else:
            threshold = self.metadata_[step]["threshold"]
            if self.threshold_type in ["dynamic", "absolute"]:
                threshold_idx = self.metadata_[step]["threshold_index"]
            if population is None:
                population = self.metadata_[step]["population"]

        # Output should consider threshold and return a single value per
        # observation
        probs = self.model_.predict_proba(population.data)[:, 1]
        if self.threshold_type in ["dynamic", "absolute"]:
            pred = np.zeros(probs.shape, dtype=int)
            idx = np.argsort(probs)[threshold_idx:]
            pred[idx] = 1
        else:
            pred = (probs >= threshold).astype(int)

        if return_scores:
            return (
                pd.Series(pred, index=population.data.index),
                pd.Series(probs, index=population.data.index),
            )
        else:
            return pd.Series(pred, index=population.data.index)

    def counterfactual(self, population=None, step=None):
        """
        Get the counterfactual examples.
        """
        if population is None:
            population = (
                self.population_ if step is None else self.metadata_[step]["population"]
            )

        # Ensure threshold is up-to-date
        threshold = (
            self.threshold_ if step is None else self.metadata_[step]["threshold"]
        )
        if np.isnan(threshold):
            return population.data
        else:
            self.recourse.threshold = threshold
            return self.recourse.counterfactual(population)

    def _counterfactual_gaussian_vectors(self, factuals, counterfactuals):
        score_change = np.abs(self._rng.normal(loc=0, scale=self.adaptation_))
        curr_scores = self.model_.predict_proba(self.population_.data)[:, 1]
        target_scores = np.clip(curr_scores + score_change, 0, 0.999)

        # Get base vectors
        base_vectors = counterfactuals - factuals
        base_dist = np.linalg.norm(base_vectors, axis=1).reshape(-1, 1)
        base_vectors = base_vectors / base_dist

        # Get updated intersect
        intercept = self.model_.intercept_ - np.log(target_scores / (1 - target_scores))

        # Get multiplier
        multiplier = (
            -intercept - np.dot(factuals, self.model_.coef_.T).squeeze()
        ) / np.dot(base_vectors, self.model_.coef_.T).squeeze()

        cf_vectors = multiplier.reshape(-1, 1) * base_vectors
        cf_vectors.loc[base_dist == 0] = 0
        return cf_vectors

    def _counterfactual_stepwise_vectors(self, factuals, counterfactuals):
        """Used only for stepwise and gaussian adaptation approaches."""

        # Compute avg distance to counterfactual
        base_dist = np.linalg.norm(counterfactuals - factuals, axis=1).reshape(-1, 1)
        avg_dist = base_dist[~np.isnan(base_dist)].mean()

        # Get base counterfactual vectors
        cf_vectors = avg_dist * ((counterfactuals - factuals) / base_dist)

        # Compute adaptation ratio
        # Compute euclidean distance for each agent
        return cf_vectors

    def _save_metadata(self):
        """
        Store metadata for each step. Currently incomplete.
        """

        if not hasattr(self, "metadata_"):
            self.metadata_ = {}

        self.metadata_[self.step_] = {
            "population": deepcopy(self.population_),
            "adaptation": deepcopy(self.adaptation_),
            "outcome": self.outcome_,
            "score": self.scores_,
            "threshold": self.threshold_,
            "growth_k": self.growth_k_,
        }
        if self.threshold_type in ["dynamic", "absolute"]:
            self.metadata_[self.step_]["threshold_index"] = self.threshold_index_

    @abstractmethod
    def add_agents(self, n_agents):
        # Generate data
        # Add info to population
        pass

    def remove_agents(self, mask):
        indices = np.where(~mask.astype(bool))[0]
        self.population_.data = self.population_.data.iloc[indices]
        self.adaptation_ = self.adaptation_.iloc[indices]
        return self

    def adapt_agents(self, population):
        """
        Applies changes to the population according to the adaptation type chosen.

        Returns the data for the adapted individuals, the original data, and the
        counterfactuals.
        """
        # Get factuals + counterfactuals
        factuals = self.population_.data
        counterfactuals = self.counterfactual()

        # Get adaptation rate for all agents
        if self.adaptation_type in ["binary", "binary_fixed"]:
            cf_vectors = counterfactuals - factuals

        elif self.adaptation_type == "stepwise":
            cf_vectors = self._counterfactual_stepwise_vectors(
                factuals, counterfactuals
            )

        elif self.adaptation_type == "gaussian":
            cf_vectors = self._counterfactual_gaussian_vectors(
                factuals, counterfactuals
            )
            new_factuals = factuals + cf_vectors
            new_factuals = np.clip(
                new_factuals,
                self.population_.action_set_.lb,
                self.population_.action_set_.ub,
            )
            return new_factuals, factuals, counterfactuals

        else:
            raise NotImplementedError()

        # Update existing agents' feature values
        new_factuals = factuals + self.adaptation_.values.reshape(-1, 1) * cf_vectors
        new_factuals = np.clip(
            new_factuals,
            self.population_.action_set_.lb,
            self.population_.action_set_.ub,
        )
        return new_factuals, factuals, counterfactuals

    def update(self):
        """
        Moves environment to the next time step.

        Updates threshold, agents in population, model (?)
        The update must:
        - Remove the Agents with favorable outcomes
        - Add new Agents
        - Remove Agents with adverse outcomes that would give up
        """

        # Remove agents with favorable outcome from previous step
        if self.remove_winners:
            self.remove_agents(self.outcome_)

        # Adapt agents
        new_factuals, factuals, counterfactuals = self.adapt_agents(
            population=self.population_.data
        )
        self.population_.data = new_factuals

        # Add new agents
        self._new_agents = self.add_agents(self.growth_k_)
        self._new_agents.index = range(
            self._max_id + 1, self._max_id + self._new_agents.shape[0] + 1
        )
        self.population_.data = pd.concat([self.population_.data, self._new_agents])
        self._max_id = self.population_.data.index.max()

        # Update variables
        self.growth_k_ = self._update_growth_rate()
        self.threshold_ = self._update_threshold()
        self.outcome_, self.scores_ = self.predict(return_scores=True)
        self.adaptation_ = self._update_adaptation()

        # Update metadata and step number
        self.step_ += 1
        self._save_metadata()

        return self

    def run_simulation(self, n_steps):
        """
        Simulates ``n`` runs through the environment.
        """
        for _ in range(n_steps):
            self.update()
        return self

    def _get_moving_agents(self, step):
        """Get indices of agents that adapted between ``step-1`` and ``step``."""
        if step == 0:
            raise IndexError("Agents cannot move at the initial state (``step=0``).")

        adapted = (
            (self.metadata_[step - 1]["adaptation"] > 0)
            & (~self.metadata_[step - 1]["outcome"].astype(bool))
            & (
                self.model_.predict_proba(self.metadata_[step - 1]["population"].data)[
                    :, 1
                ]
                < self.metadata_[step - 1]["threshold"]
            )
        )
        return adapted[adapted].index.values

    # NOTE: THIS SECTION ONWARDS IS DEDICATED TO ANALYSIS.
    #       THESE METHODS SHOULD BE IN THEIR OWN SUBMODULE.
    def success_rate(self, step, last_step=None):
        """
        For an agent to move, they need to have adaptation > 0, unfavorable outcome and
        score < threshold.

        If nan, no agents adapted (thus no rate is calculated).

        NOTE: DEPRECATED; REMOVE SOON.
        """
        if step == 0:
            raise IndexError(
                "Cannot calculate success rate at ``step=0`` (agents have not adapted "
                "yet)."
            )

        steps = [step] if last_step is None else [s for s in range(step, last_step)]
        success_rates = []
        for step in steps:
            # Indices of agents that moved
            adapted = self._get_moving_agents(step)

            # Indices of agents above threshold (according to the last state)
            above_threshold = (
                self.metadata_[step]["score"] >= self.metadata_[step - 1]["threshold"]
            )
            above_threshold = above_threshold[above_threshold].index.to_numpy()

            # Indices of agents that moved and crossed the threshold
            candidates = np.intersect1d(adapted, above_threshold)

            # Indices of all agents with favorable outcome
            favorable_outcomes = self.metadata_[step]["outcome"].astype(bool)
            favorable_outcomes = favorable_outcomes[favorable_outcomes].index.to_numpy()

            # Indices of agents that moved and have a favorable outcome
            success = np.intersect1d(favorable_outcomes, candidates)
            success_rate = (
                success.shape[0] / candidates.shape[0]
                if candidates.shape[0] > 0
                else np.nan
            )

            success_rates.append(success_rate)
        return np.array(success_rates)

    def threshold_drift(self, step=None, last_step=None):
        """
        Calculate threshold variations across time steps.

        NOTE: DEPRECATED; REMOVE SOON.
        """

        if step == 0:
            raise IndexError("Cannot get threshold drift at ``step=0``.")

        step = 1 if step is None else step
        last_step = max(self.metadata_.keys()) if last_step is None else last_step

        steps = [step] if last_step is None else [s for s in range(step, last_step + 1)]
        steps = [step - 1] + steps
        thresholds = [self.metadata_[step]["threshold"] for step in steps]
        threshold_drift = [
            (thresholds[i] - thresholds[i - 1]) / thresholds[i - 1]
            for i in range(1, len(thresholds))
        ]
        return np.array(threshold_drift)

    def agents_info(self):
        """
        Return a dataframe with number of adaptations performed by each agent that
        entered the environment.

        Contains the following features:
        - n_adaptations: number of times the agent adapted.
        - entered_step: step in which the agent entered the environment.
        - favorable_step: step in which the agent obtained a positive outcome. If nan,
          the agent never received a positive outcome.
        - original_score: score of the agent upon entering the environment.
        - final_score: score of the agent when obtaining the positive outcome.
        - n_failures: number of times agent adapted, crossed the threshold but didn't
          obtain a positive outcome.
        """
        idx = {
            step: meta["population"].data.index for step, meta in self.metadata_.items()
        }

        # entered_step
        entered = [
            pd.Series(i, index=idx[i][~idx[i].isin(idx[i - 1])])
            if i != 0
            else pd.Series(i, index=idx[i])
            for i in idx.keys()
        ]
        df = pd.concat(entered).to_frame("entered_step")

        # n_adaptations
        moving = [self._get_moving_agents(i) for i in idx.keys() if i != 0]
        moving = pd.Series(
            Counter([i for submoving in moving for i in submoving]),
            name="n_adaptations",
        )
        df = pd.concat([df, moving], axis=1).copy()
        df["n_adaptations"] = df["n_adaptations"].fillna(0).astype(int)

        # favorable_step
        favorable = [(i, self.metadata_[i]["outcome"]) for i in idx.keys()]
        favorable = pd.concat(
            [
                pd.Series(i, index=outcome[outcome == 1].index, name="favorable_step")
                for i, outcome in favorable
            ]
        )
        df = pd.concat([df, favorable], axis=1)

        # original_score
        df["original_score"] = df.apply(
            lambda row: self.metadata_[row["entered_step"]]["score"].loc[row.name],
            axis=1,
        )

        # final_score
        df["final_score"] = df.apply(
            lambda row: (
                self.metadata_[row["favorable_step"]]["score"].loc[row.name]
                if not np.isnan(row["favorable_step"])
                else np.nan
            ),
            axis=1,
        )

        # n_failures
        df["n_failures"] = 0
        for step in idx.keys():
            if step == 0:
                continue
            adapted = self._get_moving_agents(step)
            above_threshold = (
                self.metadata_[step]["score"] >= self.metadata_[step - 1]["threshold"]
            )
            above_threshold = above_threshold[above_threshold].index.to_numpy()
            candidates = np.intersect1d(adapted, above_threshold)
            unfavorable = ~self.metadata_[step]["outcome"].astype(bool)
            unfavorable = unfavorable[unfavorable].index.to_numpy()
            failed = np.intersect1d(unfavorable, candidates)
            df.loc[failed, "n_failures"] += 1
        return df

    def steps_info(self):
        """
        Return a dataframe with information regarding each time step in the environment.

        Contains the following features:
        - n_adapted: number of agents that adapted.
        - n_candidates: number of agents that adapted and crossed the previous time
          step's threshold.
        - favorable_outcomes: number of favorable outcomes.
        - success_rate: percentage of favorable outcomes within agents that adapted and
          crossed the threshold.
        - threshold: threshold value.
        - threshold_drift: percentage change between thresholds.
        - new_agents: number of new agents
        - new_agents_proba: probability of a single new agent to be above the threshold.
        - moving_agent_proba: mean likelihood of an agent to adapt and cross the
          threshold.
        - success_proba: probability of an agent adapting towards its counterfactual to
          achieve a positive outcome.
        """
        info = {}
        for step in self.metadata_.keys():
            if step == 0:
                continue

            # Number of agents that moved
            adapted = self._get_moving_agents(step)

            # Number of agents that moved and crossed the threshold
            above_threshold = (
                self.metadata_[step]["score"] >= self.metadata_[step - 1]["threshold"]
            )
            above_threshold = above_threshold[above_threshold].index.to_numpy()
            candidates = np.intersect1d(adapted, above_threshold)

            # Number of agents with favorable outcome
            favorable_outcomes = self.metadata_[step]["outcome"].astype(bool)
            favorable_outcomes = favorable_outcomes[favorable_outcomes].index.to_numpy()

            # Indices of agents that moved and have a favorable outcome
            success = np.intersect1d(favorable_outcomes, candidates)
            success_rate = (
                success.shape[0] / candidates.shape[0]
                if candidates.shape[0] > 0
                else np.nan
            )

            # Calculate threshold drift
            threshold_prev = self.metadata_[step - 1]["threshold"]
            threshold = self.metadata_[step]["threshold"]
            threshold_drift = (threshold - threshold_prev) / threshold_prev

            # Number of new agents
            idx_prev = self.metadata_[step - 1]["population"].data.index
            idx = self.metadata_[step]["population"].data.index
            new_agents = idx[~idx.isin(idx_prev)].shape[0]

            # Probability of achieving a favorable outcome

            # Create entry for dataframe
            info[step] = {
                "n_adapted": adapted.shape[0],
                "n_candidates": candidates.shape[0],
                "favorable_outcomes": favorable_outcomes.shape[0],
                "success_rate": success_rate,
                "threshold": threshold,
                "threshold_drift": threshold_drift,
                "new_agents": new_agents,
                "new_agents_proba": self.new_agent_proba(threshold),
                "moving_agent_proba": self.moving_agent_proba(step=step).mean(),
                "success_proba": self.success_proba(step=step),
            }
        return pd.DataFrame(info).T

    def new_agent_proba(self, threshold):
        """
        Calculates the probability for a new agent to be above a given threshold,
        based on the distribution of previously added agents.
        """
        scores = self.metadata_[0]["score"]
        return (scores >= threshold).astype(int).sum() / scores.shape[0]

    def n_new_agents_proba(self, n_above, new_agents=None, threshold=None, step=None):
        """
        Calculates the probability for at least ``n`` new agents to be above a given
        threshold within a set of newly-introduced agents, based on the distribution of
        previously added agents.
        """
        step = self.step_ if step is None else step

        if threshold is None:
            threshold = self.metadata_[step]["threshold"]

        if new_agents is None:
            new_agents = self.metadata_[step]["growth_k"]

        p = self.new_agent_proba(threshold)
        dist = binom(new_agents, p)
        return sum([dist.pmf(i) for i in range(n_above, new_agents + 1)])

    def moving_agent_proba(self, step=None):
        """
        Calculate probability of each agent to simultaneously adapt and cross the
        threshold.
        """
        if step is None:
            step = self.step_

        scores = self.metadata_[step]["score"]
        threshold = self.metadata_[step]["threshold"]
        adaptation = self.metadata_[step]["adaptation"]

        mask = scores < threshold
        scores = scores[mask]
        adaptation = adaptation[mask]

        if self.adaptation_type in ["binary", "binary_fixed"]:
            p = adaptation.sum() / adaptation.shape[0]
        elif self.adaptation_type == "gaussian":
            p = norm(loc=0, scale=adaptation).sf(threshold - scores) * 2
        elif self.adaptation_type == "stepwise":
            raise NotImplementedError()

        return pd.Series(p, index=scores.index)

    def n_moving_agents_proba(self, n_above, step=None):
        """
        Calculates the probability for at least ``n`` agents to move above a given
        threshold within the set of agents in the environment.

        NOTE: Continuity correction is applied, i.e., the output corresponds to
        ``P(X >= n_above-0.5)``.
        """

        if step is None:
            step = self.step_

        p = self.moving_agent_proba(step)
        mean = p.sum()
        std = np.sqrt(np.sum(p * (1 - p)))

        # Applying continuity correction
        return norm(loc=mean, scale=std).sf(n_above - 0.5)

    @np.vectorize
    def _success_single_comb(self, new, ada, step):
        p_new = self.n_new_agents_proba(new, step) - self.n_new_agents_proba(
            new + 1, step
        )
        p_ada = self.n_moving_agents_proba(ada, step) - self.n_moving_agents_proba(
            ada + 1, step
        )
        return p_new * p_ada

    def success_proba(self, step=None):
        """
        Calculate the probability of an agent adapting exactly towards its counterfactual
        to receive a positive outcome, i.e., calculates
        ``P(pos_outcome | delta_score = threshold - score)``.
        """
        if step is None:
            step = self.step_

        scores = self.metadata_[step]["score"]
        threshold = self.metadata_[step]["threshold"]
        n_pos = self.metadata_[step]["outcome"].sum()

        # Get all combinations of numbers of negative outcomes
        new_agents = list(range(self.metadata_[step]["growth_k"] + 1))
        adapting_agents = list(range((scores < threshold).sum() + 1))
        combinations = np.array(
            [(i, j) for i, j in product(new_agents, adapting_agents) if i + j < n_pos]
        )

        return self._success_single_comb(self, *combinations.T, step).sum()
